
""" wfGenes: Automatic workflow generator."""


__author__ = 'Mehdi Roozmeh'
__email__ = 'mehdi.roozmeh@kit.edu'
__copyright__ = 'Copyright 2020, Karlsruhe Institute of Technology'

import os
import os.path
import sys
from collections import OrderedDict
import json
import yaml
from wfgenes.core.initial_wfgenes import BasewfGenes
import time

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))


class DotwfGenes():
    """ DOT file generation from a wconfig file """

    def __init__(self,  wfgenes_init):
        self.dot_generation(wfgenes_init)

    def dot_generation(self, wfgenes_init):
        wfgenes_init.dot_path = os.path.join(wfgenes_init.workflow_path, 'DOT')
        def_color = 'none'
        def_edgecolor = ''
        if not os.path.exists(wfgenes_init.dot_path):
            os.makedirs(wfgenes_init.dot_path)
        dot_file = os.path.join(
            wfgenes_init.dot_path, wfgenes_init.interface_dict['workflow_name'] + '.dot')
        dot_string = 'digraph {\n'
        dot_string += '  graph[\n'
        dot_string += '\tname=' + wfgenes_init.interface_dict['workflow_name'] + '\n'
        dot_string += '  ]; \n'
        for i in range(wfgenes_init.routine_number):
            dot_string += str(i) + ' [' + '\n'
            dot_string += 'state=NONE \n'
            dot_string += 'name="' + wfgenes_init.routine_name[i] + '"\n'
            dot_string += 'label="' + wfgenes_init.routine_name[i] + '"\n'
            dot_string += ' style=filled \n'
            dot_string += 'color="' + def_edgecolor + '"\n'
            dot_string += 'fillcolor="'+def_color+'"\n'
            dot_string += '  ]; \n'

        for i in range(wfgenes_init.routine_number):
            for j in range(wfgenes_init.subroutine_number[i]):
                for k in range(len(wfgenes_init.inputs[i][j])):
                    if wfgenes_init.inputs_nodup[i][j][k] != \
                        'duplicate' and wfgenes_init.inputs_no_locdep[i][j][k] != 'dependent'\
                            and wfgenes_init.inputs_links[i][j][k][0].isnumeric() and wfgenes_init.inputs_locname[i][j][k] != 'NULL':
                        dot_string += wfgenes_init.inputs_links[i][j][k][0] + \
                            '->' + str(i) + '[ \n'
                        dot_string += 'label="' + \
                            wfgenes_init.inputs_locname[i][j][k] + '"\n'
                        dot_string += '  ]; \n'

        dot_string += '  } \n'

        with open(dot_file, 'w') as file:
            file.write(dot_string)
        self.pdf_file = os.path.join(
            wfgenes_init.dot_path, wfgenes_init.interface_dict['workflow_name'] + '.pdf')
        self.png_file = os.path.join(
            wfgenes_init.dot_path, wfgenes_init.interface_dict['workflow_name'] + '.png')

        cmd = 'dot -Tpdf ' + dot_file + ' -o ' + self.pdf_file
        os.popen(cmd)
        cmd = 'dot -Tpng ' + dot_file + ' -o ' + self.png_file
        os.popen(cmd)

        



from igraph import Graph
import numpy as np

def get_resistance_distance(self, normalized=False):
    weights=None
    if self.is_weighted():
        weights='weight'
    L = np.array(self.laplacian(weights=weights, normalized=normalized))
    L_plus = np.linalg.pinv(L)
    shape = L_plus.shape
    Omega = np.empty(shape)
    for i in range(shape[0]):
        for j in range(shape[1]):
            Omega[i][j] = L_plus[i][i]+L_plus[j][j]-L_plus[i][j]-L_plus[j][i]
    return Omega


def get_admittance_distance(self, normalized=False):
    Omega = self.get_resistance_distance(normalized=normalized)
    shape = Omega.shape
    Lambda = np.zeros(shape)
    for i in range(shape[0]):
        for j in range(shape[1]):
            if i != j: Lambda[i][j] = 1./Omega[i][j]
    return Lambda


def get_kirchhoff_index(self, mode='resistance', normalized=False):
    if mode is 'resistance':
        matrix = self.get_resistance_distance()
    elif mode is 'admittance':
        matrix = self.get_admittance_distance()
    else:
        raise ValueError('invalid mode: ' + mode)
    if normalized:
        shape = matrix.shape
        return np.sum(matrix, axis=(0,1))/(shape[0]*shape[1])
    else:
        return np.sum(matrix, axis=(0,1))/2.


def get_spectrum(self, mode='laplacian'):
    if mode is 'laplacian':
        return np.sort(
            np.linalg.eigvals(self.laplacian(normalized=True))
            )
    elif mode is 'adjacency':
        if self.is_weighted():
            attribute = 'weight'
        else:
            attribute = None
        return np.sort(
            np.linalg.eigvals(self.get_adjacency(attribute=attribute).data)
            )
    else:
        raise ValueError('invalid mode: ' + mode)

Graph.get_kirchhoff_index = get_kirchhoff_index
Graph.get_resistance_distance = get_resistance_distance
Graph.get_admittance_distance = get_admittance_distance
Graph.get_spectrum = get_spectrum

"""
# generic method tests
#gr = Graph.Full(3)
#gr = Graph.Star(8)
#gr = Graph.Ring(6)
# gr = Graph.Full_Bipartite(4, 4)
# gr = Graph.Famous('Tetrahedral')
gr = Graph.Famous('Octahedral')
# gr = Graph.Famous('Cubical')
# gr = Graph.Formula('A-->B-->C-->D-->A')

print gr.get_resistance_distance()
print gr.get_kirchhoff_index()
print gr.get_spectrum()
"""

""" http://stackoverflow.com/questions/31034730/graph-analysis-identify-loop-paths """
def cycle_basis(G, root=None):
    gnodes = set(n.index for n in G.vs())
    cycles = []
    while gnodes: # loop over connected components
        if root is None:
            root = gnodes.pop()
        stack = [root]
        pred = {root:root} 
        used = {root:set()}
        while stack: # walk the spanning tree finding cycles
            z = stack.pop() # use last-in so cycles easier to find
            zused = used[z]
            for nbr in G.neighbors(z, mode='ALL'):
                if nbr not in used: # new node 
                    pred[nbr] = z
                    stack.append(nbr)
                    used[nbr] = set([z])
                elif nbr is z: # self loops
                    cycles.append([z]) 
                elif nbr not in zused: # found a cycle
                    pn = used[nbr]
                    cycle = [nbr,z]
                    p = pred[z]
                    while p not in pn:
                        cycle.append(p)
                        p = pred[p]
                    cycle.append(p)
                    cycles.append(cycle)
                    used[nbr].add(z)
        gnodes -= set(pred)
        root = None
    return cycles

"""
g = Graph()
g.add_vertices(9)
g.add_edges(
    [
    (0,1), (1,2), (2,0), (2,3), (3,0), (1,4), (4,5), 
    (5,2), (0,5), (0,6), (6,7), (7,8), (8,5)
    ]
)
print('cycle_basis_ig: ', cycle_basis(g))
"""




""" functions for constructing, manipulating and analyzing morphologies """
__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2016, Karlsruhe Institute of Technology'


import random as rnd
import numpy as np
from ase import Atoms
from ase.data.vdw import vdw_radii
from ase.geometry import find_mic


def vdw_clash(morphology, num):
    """ Test whether a molecule (particle) is clashing with some of the rest """
    cutoffs = [vdw_radii[number] for number in morphology.get_atomic_numbers()]
    mol_new_ind = range(len(morphology)-num, len(morphology))
    rest_ind = range(0, len(morphology)-num)
    for index1 in mol_new_ind:
        distances = morphology.get_distances(index1, rest_ind, mic=True)
        for index2, distance in enumerate(distances):
            if distance < cutoffs[index1]+cutoffs[index2]:
                return True
    return False


def rmsd(reference, structure):
    """ Calculates the Root Mean Square Deviation between a structure
    and a reference both provided as Atoms objects """
    from ase.build import minimize_rotation_and_translation
    var_structure = Atoms(structure)
    minimize_rotation_and_translation(reference, var_structure)
    return np.sqrt(
        np.mean(
            (np.linalg.norm(var_structure.get_positions()
                            -reference.get_positions(), axis=1))**2, axis=0
        )
    )


def deposit(mol, parameters):
    """ Deposition of molecules of one type in a box. The molecules get random
    positions and orientations and clashing molecules are removed """
    clash_distance = parameters['clash distance']
    trials = parameters['trials']
    seed = parameters['rng seed']
    a = parameters['cell size']
    thkz = 1.0
    varz = True

    mol.set_cell([a, a, a])
    mol.set_pbc((1, 1, 1))
    pbc = mol.get_pbc()
    cell = mol.get_cell()
    natoms = len(mol)
    if varz:
        depz = 0.0
        mol.center(about=([a/2, a/2, 0.]))
    else:
        depz = thkz
        mol.center(about=([a/2, a/2, a/2]))

    inserted = 1
    clashed = 0
    depz_step = thkz/trials

    morphology_list = [mol.copy()]
    morphology_stack = mol.copy()

    rnd.seed(seed)

    trial = 0
    while trial < trials:
        trial += 1
        mol_new = mol.copy()
        if varz:
            mol_new.translate(
                [
                    rnd.uniform(-a/2, a/2),
                    rnd.uniform(-a/2, a/2),
                    rnd.uniform(max(0., a*depz-clash_distance), a*depz)
                ]
            )
            depz = depz + depz_step
        else:
            mol_new.translate(
                [
                    rnd.uniform(-a/2, a/2),
                    rnd.uniform(-a/2, a/2),
                    rnd.uniform(-a/2*depz, a/2*depz)
                ]
            )
        mol_new.euler_rotate(
            phi=rnd.uniform(-180., 180.),
            theta=rnd.uniform(-90., 90.),
            psi=rnd.uniform(-180., 180.),
            center='COM'
        )
        mol_new_com = mol_new.get_center_of_mass()
        distances = []
        for molecule in morphology_list:
            distlist = [mol_new_com-molecule.get_center_of_mass()]
            distances.append(find_mic(distlist, cell, pbc)[1].tolist()[0])
        if all(distance > clash_distance for distance in distances):
            morphology_stack += mol_new
            if vdw_clash(morphology_stack, natoms):
                # remove the just added molecule
                del morphology_stack[-natoms:]
                clashed += 1
            else:
                morphology_list.append(mol_new)
                inserted += 1
        else:
            clashed += 1

    return morphology_list


def deposit_random(mol, parameters):
    """ Deposition of molecules of one type in a box. The molecules get random
    positions and orientations and are all accepted """
    trials = parameters['trials']
    a = parameters['lattice parameter']

    mol.set_cell([a, a, a])
    mol.set_pbc((1, 1, 1))
    mol.center(about=([a/2, a/2, a/2]))

    morphology_list = [mol.copy()]

    rnd.seed(1)

    trial = 0
    while trial < trials:
        trial += 1
        mol_new = mol.copy()
        mol_new.translate(
            [
                rnd.uniform(-a/2, a/2),
                rnd.uniform(-a/2, a/2),
                rnd.uniform(-a/2, a/2)
            ]
        )
        mol_new.euler_rotate(
            phi=rnd.uniform(-180., 180.),
            theta=rnd.uniform(-90., 90.),
            psi=rnd.uniform(-180., 180.),
            center='COM'
        )
        morphology_list.append(mol_new)

    return morphology_list


def gyration_radius(positions):
    """ calculate radius of gyration of particles with provided positions """
    from math import sqrt
    centroid = np.mean(positions, axis=0)
    rog = sqrt(np.mean((np.linalg.norm(positions-centroid, axis=1))**2, axis=0))
    return rog




""" wfGenes: Automatic workflow generator."""


__author__ = 'Mehdi Roozmeh'
__email__ = 'mehdi.roozmeh@kit.edu'
__copyright__ = 'Copyright 2020, Karlsruhe Institute of Technology'

import os
import os.path
from copy import deepcopy
from collections import OrderedDict
import shutil
import argparse
import json
import xmlschema
import yaml
from wfgenes.core.initial_wfgenes import BasewfGenes



class CleanwfGenes(BasewfGenes):
    """"Clean up previous files for old wConfig"""
    def __init__(self, blueargs):
        #BasewfGenes.__init__(self, blueargs)
        self.clean_wfgenes()


    def clean_wfgenes(self):
        if os.path.exists(self.workflow_path):
            shutil.rmtree(self.workflow_path)   


""" Functions for modeling charge transport in organic semiconductors for
    parallel workflow execution
"""
__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2016, Karlsruhe Institute of Technology'


def prepare_input(parameters):
    """
    input: parameters file
    output: parameters, monomer_structure
    """
    from ase.io import read
    from util import atoms2dict

    title = 'prepare input'
    print('Processing step:', title)

    monomer = read(str(parameters['monomer file']))
    monomer_structure = atoms2dict(monomer, plain_arrays=True)

    return monomer_structure, parameters


def optimize_geometry(monomer_structure, parameters):
    """
    node input: monomer structure, parameters
    node output: optimized monomer structure
    """
    from util import dict2atoms, atoms2dict
    from calculators import calculate

    monomer = dict2atoms(monomer_structure, plain_arrays=True)

    title = 'geometry optimization'
    print('Processing step:', title)

    loc_parameters = parameters['calculator']['parameters'].copy()
    loc_parameters['task'] = 'optimize'
    calculate(monomer, parameters['calculator']['name'], loc_parameters)
    return atoms2dict(monomer, plain_arrays=True)


def reorganization_energy(monomer_structure, parameters):
    """
    node input: monomer structure, parameters (ro)
    node output: reorganization_energy
    """
    from util import dict2atoms
    from transport import reorg_energy

#    molecule = dict2atoms(monomer_structure, plain_arrays=True)
#    properties = {}

    title = 'reorganization energy'
    print('Processing step:', title)

    if parameters['carrier'] == 'electron':
        parameters['ion charge'] = parameters['charge']-1
    if parameters['carrier'] == 'hole':
        parameters['ion charge'] = parameters['charge']+1

    if 'reorganization energy' in parameters:
        lambda_t = parameters['reorganization energy']
        print('reorganization energy read from input')
    else:
        lambda_t = reorg_energy(
            structure=dict2atoms(monomer_structure, plain_arrays=True),
            parameters=parameters
        )
#    properties['reorganization energy'] = {'value': lambda_t, 'units': 'eV'}
#    return properties['reorganization energy']
    return {'value': lambda_t, 'units': 'eV'}


def generate_morphology(monomer_structure, parameters, properties):
    """
    input: monomer structure, parameters
    output: morphology, properties
    """
    import matplotlib
    matplotlib.use('Agg') # Force matplotlib to not use any Xwindows backend
    import matplotlib.pyplot as plt
    from ase import Atoms
    from ase.io import write
    from ase.units import _amu
    from util import dict2atoms, atoms2dict
    from molecules import Morphology
    from morphology import deposit

    monomer = dict2atoms(monomer_structure, plain_arrays=True)

    title = 'generate morphology'
    print('Processing step:', title)
    morphology_type = parameters['morphology type']
    morphology_scale = parameters['morphology scale']
    if morphology_scale == 'atomistic':
        monomer = dict2atoms(monomer_structure, plain_arrays=True)
    if morphology_type == 'amorphous' and morphology_scale == 'atomistic':
        molecules = deposit(monomer, parameters)
        morphology = Morphology.from_molecules(molecules)
    elif morphology_type == 'crystalline' and morphology_scale == 'coarse-grained':
        if parameters['crystal lattice'] == 'simple cubic':
            morphology = Morphology.SimpleCubic(
                parameters['lattice parameter'],
                parameters['supercell size']
            )
    else:
        raise RuntimeError('unknown morphology type')

    # morphology analysis
    volume = morphology.get_volume()
    mass = morphology.get_mass()
    if mass is not None:
        density_gccm = (mass/volume)*(_amu*1000*1.e24) # 1 cm^3 = 1.e24 Ang^3, _amu in kg
        properties['density'] = {'units': 'g/cm^3', 'value': density_gccm}
    [rdf, dist] = morphology.get_rdf(
        parameters['rdf cutoff radius'],
        parameters['rdf number of bins']
    )
    properties['rdf'] = [dist.tolist(), rdf.tolist()]

    # plot the RDF
    plt.plot(dist, rdf)
    plt.title(parameters['title'])
    plt.xlabel('r, Ang')
    plt.ylabel('g(r)')
    plt.savefig('morphology-rdf.png')

    # visualize morphology
    if morphology_scale == 'atomistic':
        structure = Atoms()
        for molecule in molecules:
            structure += molecule
        write('morphology-structure.xyz', structure)
        write('morphology-structure.png', structure)
        write('morphology-structure.html', structure)

    # output
    morphology_dct = morphology.todict()
    if morphology_scale == 'atomistic':
        m_array = []
        for molecule in molecules:
            m_array.append(atoms2dict(molecule, plain_arrays=True))
        morphology_dct['molecules'] = m_array
    properties['number of molecules'] = len(morphology)

    return morphology_dct, properties


def find_pairs(morphology, parameters):
    """
    input: morphology, parameters
    output: hopping sites, hopping pairs
    """
    from util import dict2atoms
    from molecules import Morphology
    from transport import get_pairs

    if 'molecules' in morphology:
        molecules = []
        for molecule in morphology['molecules']:
            molecules.append(dict2atoms(molecule, plain_arrays=True))
        morphology = Morphology.from_molecules(molecules)
    else:
        morphology = Morphology(**morphology)

    assert len(morphology) > 1, 'too few or no molecules in morphology'

    title = 'find pairs'
    print('Processing step:', title)

    [hopping_sites, hopping_pairs] = get_pairs(morphology, parameters['hopping threshold'])

    assert len(hopping_sites) > 0, 'no hopping sites found'

    return hopping_sites, hopping_pairs


def site_energies(hopping_sites, morphology, parameters):
    """
    input: hopping sites, morphology, parameters
    output: hopping sites
    """
    import numpy as np
    from util import dict2atoms
    from calculators import get_energy

    title = 'site energies'
    print('Processing step:', title)

    molecules = []
    for molecule in morphology['molecules']:
        molecules.append(dict2atoms(molecule, plain_arrays=True))

    assert len(hopping_sites) > 0, 'list hopping_sites has zero length'

    if parameters['carrier'] == 'electron':
        parameters['ion charge'] = parameters['charge']-1
    if parameters['carrier'] == 'hole':
        parameters['ion charge'] = parameters['charge']+1

    if not isinstance(hopping_sites, list):
        singletask = True
        hopping_sites = [hopping_sites]
    else:
        singletask = False

    calc_params = parameters['calculator']['parameters']
    calc_name = parameters['calculator']['name']
    for site in hopping_sites:
        site_atoms = molecules[site['index']]
        site['energy'] = get_energy(site_atoms, calc_name, calc_params)
        if 'electric field' in parameters:
            q = parameters['ion charge']-parameters['charge']
            site['energy'] += -q*np.dot(
                parameters['electric field'],
                site['center of mass']
            )
        calc = site_atoms.get_calculator()
        eigenvectors = calc.get_mos()[0]
        [number, energy] = calc.get_mo('HOMO')
        site['HOMO'] = {
            'energy': energy,
            'number': number,
            'eigenvector': eigenvectors[number-1].tolist()
        }
        [number, energy] = calc.get_mo('LUMO')
        site['LUMO'] = {
            'energy': energy,
            'number': number,
            'eigenvector': eigenvectors[number-1].tolist()
        }

    if singletask:
        return hopping_sites[0]
    else:
        return hopping_sites


def electronic_couplings(hopping_pairs, hopping_sites, morphology, parameters):
    """
    input: hopping pairs
    output: hopping pairs
    """
    from util import dict2atoms
    from transport import get_e_coupling

    molecules = []
    for molecule in morphology['molecules']:
        molecules.append(dict2atoms(molecule, plain_arrays=True))

    assert len(hopping_pairs) > 0, 'list hopping_pairs has zero length'

    title = 'electronic couplings'
    print('Processing step:', title)

    if not isinstance(hopping_pairs, list):
        singletask = True
        hopping_pairs = [hopping_pairs]
    else:
        singletask = False

    for pair in hopping_pairs:
        pair['coupling'] = get_e_coupling(
            left_index=pair['left'],
            right_index=pair['right'],
            parameters=parameters,
            morphology=molecules,
            hopping_sites=hopping_sites)

    if singletask:
        return hopping_pairs[0]
    else:
        return hopping_pairs


def analyses(simulation):
    """
    input: hopping sites, hopping pairs
    output: [deltaE, rate] for each hopping pair, sigma, mean_coupling, mobility
    parameters (as constants): n, C, M
    """
    from util import dict2atoms
    from ase import Atoms
    from ase.io import write
    from transport import marcus_rates, mobility
    import matplotlib
    matplotlib.use('Agg') # Force matplotlib to not use any Xwindows backend
    import matplotlib.pyplot as plt

    simulation['properties']['reorganization energy'] = simulation['reorganization energy']
    del simulation['reorganization energy']

    title = 'charge transfer rates'
    print('Processing step:', title)
    marcus_rates(simulation)

    title = 'carrier mobility'
    print('Processing step:', title)
    mobility(simulation)

    # visualization
    if 'rdf' in simulation['properties']:
        [dist, rdf] = simulation['properties']['rdf']
        plt.plot(dist, rdf)
        plt.title(simulation['parameters']['title'])
        plt.xlabel('r, Ang')
        plt.ylabel('g(r)')
        plt.savefig('morphology-rdf.png')

    # visualize morphology
    molecules = []
    for molecule in simulation['morphology']['molecules']:
        molecules.append(dict2atoms(molecule, plain_arrays=True))
    structure = Atoms()
    for molecule in molecules:
        structure += molecule
    write('morphology-structure.xyz', structure)
    write('morphology-structure.png', structure)
    write('morphology-structure.html', structure)
    write('morphology.xyz', molecules)

    return simulation



""" Extend ASE calculators with interoperability methods """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2016, Karlsruhe Institute of Technology'

from ase.units import Hartree

classmap = {}
nsmap = {}
_properties = {}
_parameters = {}

try:
    import ase.calculators.turbomole as turbomole
    from ase.calculators.turbomole import Turbomole
except ImportError:
    raise
else:
    classmap['turbomole'] = Turbomole
    nsmap['Turbomole'] = turbomole

    def get_vibrational_energies():
        """ an interface to return vibrational energies """
        pass
    Turbomole.get_vibrational_energies = get_vibrational_energies

    def get_mos(self):
        """ extracts the molecular orbitals and the orbital energies """
        import numpy as np
        if 'uhf' in self.results['calculation parameters']:
            assert not self.results['calculation parameters']['uhf']
        if 'molecular orbitals' in self.results:
            eigenvectors = []
            eigenvalues = []
            for mo in self.results['molecular orbitals']:
                if mo['spin'] is None:
                    eigenvectors.append(mo['eigenvector'])
                    eigenvalues.append(mo['eigenvalue'])
            return [np.array(eigenvectors),
                    np.diag(np.array(eigenvalues))]
    Turbomole.get_mos = get_mos

    def get_mo(self, mo_type):
        """ returns index and energy of frontier orbitals """
        assert mo_type in ['HOMO', 'LUMO']
        orb_energy = None
        orb_number = None
        if mo_type == 'HOMO':
            for orbital in self.results['molecular orbitals']:
                if orbital['occupancy'] > 0:
                    if orb_energy is None or orbital['eigenvalue'] > orb_energy:
                        orb_energy = orbital['eigenvalue']
                        orb_number = orbital['index']
        if mo_type == 'LUMO':
            for orbital in self.results['molecular orbitals']:
                if orbital['occupancy'] < 1:
                    if orb_energy is None or orbital['eigenvalue'] < orb_energy:
                        orb_energy = orbital['eigenvalue']
                        orb_number = orbital['index']
        return [orb_number, orb_energy]
    Turbomole.get_mo = get_mo

try:
    import ase.calculators.nwchem as nwchem
    from ase.calculators.nwchem import NWChem
except ImportError:
    raise
else:
    classmap.update({'nwchem': NWChem})
    nsmap['NWChem'] = nwchem
    NWChem.converged = True # temporary
    def get_mos(self):
        """ return eigenvectors and eigenvalues as matrices (numpy arrays) """
        import numpy as np
        orbital_coefficients = []
        eigen_energies = []
        for orbital in self.get_property('orbitals'):
            orbital_coefficients.append(orbital['coefficients'])
            eigen_energies.append(orbital['energy'])
        return [
            np.array(orbital_coefficients),
            np.diag(np.array(eigen_energies)*Hartree)
        ]
    NWChem.get_mos = get_mos

    def get_mo(self, mo_type):
        """ returns index and energy of frontier orbitals """
        assert mo_type in ['HOMO', 'LUMO']
        orb_energy = None
        orb_number = None
        if mo_type == 'HOMO':
            for orbital in self.results['orbitals']:
                if orbital['occupation'] > 0:
                    if orb_energy is None or orbital['energy'] > orb_energy:
                        orb_energy = orbital['energy']
                        orb_number = orbital['number']
        if mo_type == 'LUMO':
            for orbital in self.results['orbitals']:
                if orbital['occupation'] < 1:
                    if orb_energy is None or orbital['energy'] < orb_energy:
                        orb_energy = orbital['energy']
                        orb_number = orbital['number']
        return [orb_number, orb_energy*Hartree]
    NWChem.get_mo = get_mo


try:
    from ase.calculators.vasp import Vasp
except ImportError:
    raise
else:
    classmap['vasp'] = Vasp
    _properties['vasp'] = {
        'energy': {'function': 'get_potential_energy', 'args': ['atoms']},
        'forces': {'function': 'get_forces', 'args': ['atoms']},
        'dipole': {'function': 'get_dipole_moment', 'args': ['atoms']},
        'fermi': {'function': 'get_fermi_level', 'args': []},
        'stress': {'function': 'get_stress', 'args': ['atoms']},
        'magmom': {'function': 'get_magnetic_moment', 'args': ['atoms']},
        'magmoms': {'function': 'get_magnetic_moments', 'args': ['atoms']},
        'xc functional': {'function': 'get_xc_functional', 'args': []},
        'eigenvalues': {'function': 'get_eigenvalues', 'args': []},
        'occupation numbers': {'function': 'get_occupation_numbers', 'args': []},
        'number of bands': {'function': 'get_number_of_bands', 'args': []},
        'k-point weights': {'function': 'get_k_point_weights', 'args': []},
        'ibz k-points': {'function': 'get_ibz_kpoints', 'args': []},
        'number of spins': {'function': 'get_number_of_spins', 'args': []},
        'spin polarized': {'function': 'get_spin_polarized', 'args': []},
        'vibrational frequencies': {'function': 'read_vib_freq', 'args': []}
    }
    '''
    _parameters['vasp'] = [
        'float_params',
        'exp_params',
        'string_params',
        'int_params',
        'bool_params',
        'list_params',
        'special_params',
        'dict_params',
        'input_params',
        'name',
        'version',
        'niter',
        'sigma',
        'nelect',
        'converged'
    ]
    '''

    def get_vibrational_energies(self):
        """ an interface to return vibrational energies """
        import numpy as np
        (vib_ene_real, vib_ene_imag) = self.read_vib_freq() # returns in meV
        vib_ene_real = np.array(vib_ene_real)*1.0e-3
        vib_ene_imag = np.array(vib_ene_imag)*1.0e-3
        return vib_ene_real, vib_ene_imag

    Vasp.get_vibrational_energies = get_vibrational_energies


try:
    from ase.calculators.emt import EMT
except ImportError:
    raise
else:
    classmap.update({'emt': EMT})


'''
def calc_get_params(calc):
    """ Return a dictionary with all known parameters """
    params = {}
    for param in _parameters[calc.name]:
        params[param] = getattr(calc, param)
    return params
'''

def calc_get_properties(calc, req_props):
    """ Return a dictionary with properties specified in the list req_props """
    props = {}
    for prop in req_props:
        args = []
        for item in _properties[calc.name][prop]['args']:
            if hasattr(calc, item):
                args.append(getattr(calc, item))
        function = _properties[calc.name][prop]['function']
        props[prop] = getattr(calc, function)(*args)
    return props


def calc_ingest(calc_name, req_props):
    """ Ingest data from a calculation which has not been performed with ASE """
    calc = classmap[calc_name](restart=True)
    assert calc.converged
    atoms = calc.get_atoms()
    props = calc_get_properties(calc, req_props)
    dct = {'calculator': {'name': calc.name, 'parameters': calc.todict()}}
    return (atoms, props, dct)


def calc_dump(calc_name):
    """ Dump data from a calculation which has not been performed with ASE """
    from data import ddump
    calc = classmap[calc_name](restart=True)
    return ddump(calc)


def calculate(atoms, calc_name, calc_parameters):
    """ perform a calculation, return calculator and updated atoms """
    if calc_name == 'nwchem' and 'multiplicity' in calc_parameters:
        calc_parameters['mult'] = calc_parameters['multiplicity']
        del calc_parameters['multiplicity']
    if calc_name == 'nwchem' and 'total charge' in calc_parameters:
        calc_parameters['charge'] = calc_parameters['total charge']
        del calc_parameters['total charge']
    calc = classmap[calc_name](**calc_parameters)
    atoms.set_calculator(calc)
    if calc_name == 'nwchem':
        calc.atoms = atoms
        calc.calculate()
    elif calc_name in ['turbomole', 'vasp']:
        calc.calculate(atoms)
    assert calc.converged
    return calc


def get_energy(atoms, calc_name, calc_parameters):
    """ perform a calculation, return total energy and updated atoms """
    calculate(atoms, calc_name, calc_parameters)
    return atoms.get_total_energy()



import sys
import os
import os.path
from pathlib import Path
from IPython.display import Image
import ipywidgets as widgets
import json, yaml
from ipywidgets import Button, HBox, VBox, Layout
from IPython.display import HTML, display, clear_output, Markdown as md, JSON, Javascript, FileLink, Image, display_javascript, display_html, display_json
from ipyfilechooser import FileChooser
import time
from json2html import *
import wfgenes
import yaml
from wfgenes import wfgenes_gui
import site

wfgenes_path = os.path.join(site.getsitepackages()[0], 'wfgenes')

if wfgenes_gui.default_path.selected_path is not None:  
    default_path = wfgenes_gui.default_path.selected_path
else:
    default_path = wfgenes_path    

                                        
# Getting the arguments with FileChooser and SelectMultiple
config = FileChooser('')
inputp = FileChooser('')

style = {'description_width': 'initial'}

model = widgets.Dropdown(
    options=['FireWorks', 'SimStack', 'Dask', 'Parsl'],
    value='FireWorks',
    description='<b>Choose a model:<b>',
    disabled=False,
    style=style
)

model_output = widgets.Output()

embedded_wms= widgets.Checkbox(
    value=False,
    description='Embedded in FireWorks',
    disabled=False,
    indent=False
)

# Set default paths for file chooser


config.default_path= default_path
inputp.default_path = default_path


# Show folder icons for file chooser
config.use_dir_icons = True
inputp.use_dir_icons = True

# Set multiple file filter patterns for config file
config.filter_pattern = ['*.txt','*.yaml','*.json', '*.yml']

# Set titles for choosers
config.title = '<b>Select a configuration file</b>'
inputp.title = '<b>Select the input path</b>'

#Button for runing wgenerator
button = widgets.Button(description='Run wfGenes', style=style, icon='play')
output = widgets.Output()

#Button for displaying Wconfig file 
dis_wconfig = widgets.Button(description='Preview WConfig', tooltip='display wconfig', icon='eye')
output2 = widgets.Output()

#Button for editing Wconfig file 
edit_wconfig = widgets.Button(description='Edit Wconfig', tooltip='Edit configuration file for modifying the graph', icon='pen')
output2 = widgets.Output()

#Button for adding another input path
plus_input = widgets.Button(tooltip='Aadding additional input paths', icon='plus', style=style, layout=Layout(width='56px'))
output3 = widgets.Output()

#Button for saving wconfig file after modifications
save_wconfig = widgets.Button(description='save', tooltip='Save wconfig after modifications', icon='save')
updated = widgets.Output()

#Text area for displaying config file
config_textarea = widgets.Textarea(
                    placeholder='Type something',
                    description='<b>Configuration file:</b>',
                    description_tooltip = 'Edit to modify the graph',
                    disabled=False,
                    style = style,
                    layout=Layout( width='100%', height='1000px')
                 )

#Dash board for displaying the graph
graph_dash = widgets.Output()

#Prepare args to wgenerator
def prepare_args(config, input):
    args = {'workflowconfig' : config.selected, 'inputpath':  inputp.selected_path , 'wms':model.value, 'embedded': embedded_wms.value}
    return args

#Run wgenerator
def run_wgenerator(args):
    png = wfgenes.core.wgenerator.run_wfgenes(args)
    return png
    
#Display graph
def display_graph(graph):
    file = open(graph, "rb") 
    image = file.read()
    graph = widgets.Image(
            value=image,
            format='png',
            height=100,
            )
    with graph_dash:
        clear_output()
        display(graph)   
    
#Define event when run wconfig button is clicked
def run_wfgenes_clicked(b):
    with output:
        clear_output()
        if config.selected == None or inputp.selected == None:
            print('Please select a configuration file and an input path.')
            time.sleep(2)
            clear_output()
            return
        args = prepare_args(config, inputp)
        png = run_wgenerator(args)
        time.sleep(1)
        display_graph(png)

#Define event when edit config button is clicked
config_output = widgets.Output()
def on_button_clicked_2(b_2):
    with config_output:
        clear_output()
        if config.selected == None:
            print('Please select a configuration file.')
            time.sleep(1.7)
            clear_output()
            return
        config_file = open(config.selected, "r")
        config_content = config_file.read()
        config_textarea.value = config_content
        display(HBox([save_wconfig, close_wconfig]))
        display(save_msg)
        display(config_textarea)
            
#Define event when WConfig preview button is clicked
config_preview = widgets.Output()
def on_wconfig_preview_clicked(b_2):
    with config_output:
        clear_output()
        if config.selected == None:
            print('Please select a configuration file.')
            time.sleep(1.7)
            clear_output()
            return
        config_file = open(config.selected, "r")
        config_content = config_file.read()
        config_split = config.selected.split('.')
        suffix_dot = len(config_split) - 1
        if config_split[suffix_dot] == 'json': 
            config_dic = json.loads(config_content)
        elif config_split[suffix_dot] == 'yaml':
            config_dic = yaml.load(config_content, Loader=yaml.Loader)
        config_to_html = json2html.convert(json = config_dic)
        config_html = HTML(config_to_html)
        display(close_wconfig)
        display_html(config_html)
            
#Define event when save wconfig button is clicked
save_msg = widgets.Output()
def on_save_wconfig_clicked(b):
    new_content = config_textarea.value
    config_file = open(config.selected, "w")
    config_file.write(new_content)
    config_file.close()
    with save_msg:
        clear_output()
        print('Configuration file was saved.')
        time.sleep(1.5)
        clear_output()

#Button to close wconfig
close_wconfig = widgets.Button(description='close', tooltip='Close wconfig', icon='close')
def on_close_wconfig_clicked(b):
    with config_output:
        clear_output()
close_wconfig.on_click(on_close_wconfig_clicked)

#Button to clear all outputs
clear_all = widgets.Button(description='Clear', tooltip='Close graph and wconfig', icon='eraser')
def on_clear_all_clicked(b):
    with graph_dash:
        clear_output()
    with config_output:
        clear_output()
    with output:
        clear_output()
    with output2:
        clear_output()
clear_all.on_click(on_clear_all_clicked)
        
#Define event when plus input button in clicked
inputpath = []
def on_button_clicked_3(b3):
    inputpath.append(FileChooser(''))
    inputpath[len(inputpath)-1].default_path = default_path
    inputpath[len(inputpath)-1].use_dir_icons = True
    with output3:
        display(inputpath[len(inputpath)-1])
        
def model_changed(bvar):
    with model_output:
        clear_output()
        if model.value == 'Dask' or model.value == 'Parsl':  
            display(embedded_wms)
        
button.on_click(run_wfgenes_clicked)
edit_wconfig.on_click(on_button_clicked_2)
plus_input.on_click(on_button_clicked_3)
dis_wconfig.on_click(on_wconfig_preview_clicked)
save_wconfig.on_click(on_save_wconfig_clicked)
model.observe(model_changed)

#Display buttons
def display_generator():
    """"""
    display(config)
    display(inputp)
    display(plus_input, output3)
    display(model, model_output)
    display(HBox([button, clear_all]), output)
    display(HBox([ dis_wconfig, edit_wconfig]), output2, output2, graph_dash, output3, config_output)



""" Methods to construct, manipulate and analyse molecule objects """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2016, Karlsruhe Institute of Technology'

import math
import numpy as np
from ase.geometry import find_mic


class Molecules:

    _positions = None
    _masses = None
    _cell = None
    _pbc = None

    def __init__(self, **kwargs):
        if 'positions' in list(kwargs.keys()):
            self._positions = np.array(kwargs['positions'])
            self._cell = kwargs['cell']
            self._pbc = kwargs['pbc']
        if 'masses' in list(kwargs.keys()):
            self._masses = np.array(kwargs['masses'])

    @classmethod
    def from_molecules(cls, molecules):
        positions = [mol.get_center_of_mass() for mol in molecules]
        masses = [np.sum(mol.get_masses()) for mol in molecules]
        cell = molecules[0].get_cell()
        pbc = molecules[0].get_pbc()
        return cls(
            positions=positions,
            masses=masses,
            cell=cell,
            pbc=pbc
        )

    @classmethod
    def from_atoms(cls, atoms):
        return cls(
            positions=atoms.get_positions(),
            cell=atoms.get_cell(),
            pbc=atoms.get_pbc()
        )

    @classmethod
    def SimpleCubic(cls, a, size, pbc=True):
        from ase.lattice.cubic import SimpleCubic
        atoms = SimpleCubic(size=size, symbol='X', pbc=pbc, latticeconstant=a)
        return cls.from_atoms(atoms)

    @classmethod
    def FaceCenteredCubic(cls, a, size, pbc=True):
        from ase.lattice.cubic import FaceCenteredCubic
        atoms = FaceCenteredCubic(size=size, symbol='X', pbc=pbc, latticeconstant=a)
        return cls.from_atoms(atoms)

    def __len__(self):
        return len(self._positions)

    def __delitem__(self, index):
        del(self._positions[index])

    def _get_positions(self):
        return self._positions

    def get_positions(self):
        return self._positions.copy()

    def get_masses(self):
        if self._masses is not None:
            return self._masses.copy()

    def get_distance(self, a0, a1, mic=False, vector=False):
        """Return distance between two molecules with indices a0 and a1.

        Use mic=True to use the Minimum Image Convention.
        vector=True gives the distance vector (from a0 to a1).
        """

        R = np.array(self._get_positions())
        D = np.array([R[a1] - R[a0]])
        if mic:
            D, D_len = find_mic(D, self._cell, self._pbc)
        else:
            D_len = np.array([np.sqrt((D**2).sum())])
        if vector:
            return D[0]

        return D_len[0]

    def get_distances(self, a, indices, mic=False, vector=False):
        """Return distances of molecule with index a with a list of molecules.

        Use mic=True to use the Minimum Image Convention.
        vector=True gives the distance vector (from a to self[indices]).
        """

        R = np.array(self._get_positions())
        D = R[indices] - R[a]
        if mic:
            D, D_len = find_mic(D, self._cell, self._pbc)
        else:
            D_len = np.sqrt((D**2).sum(1))
        if vector:
            return D
        return D_len

    def get_all_distances(self, mic=False):
        """Return distances of all of the molecules with all of the molecules.

        Use mic=True to use the Minimum Image Convention.
        """
        L = len(self)
        R = np.array(self._get_positions())

        D = []
        for i in range(L - 1):
            D.append(R[i + 1:] - R[i])
        D = np.concatenate(D)

        if mic:
            D, D_len = find_mic(D, self._cell, self._pbc)
        else:
            D_len = np.sqrt((D**2).sum(1))

        results = np.zeros((L, L), dtype=float)
        start = 0
        for i in range(L - 1):
            results[i, i + 1:] = D_len[start:start + L - i - 1]
            start += L - i - 1
        return results + results.T

    def get_volume(self):
        """Get volume of unit cell."""
        return abs(np.linalg.det(self._cell))

    def get_pbc(self):
        return self._pbc.copy()

    def get_cell(self):
        return self._cell.copy()

    def trim(self, cutoff, directions=(False, False, True)):
        """ cut away border layers of the morphology """
        deletions = []
        for direction, select in enumerate(directions):
            if select:
                minpos = np.amin(np.array(self._positions)[:,direction])
                maxpos = np.amax(np.array(self._positions)[:,direction])
                for index, position in enumerate(self._positions):
                    in_lower = position[direction] < minpos+cutoff
                    in_upper = position[direction] > maxpos-cutoff
                    if in_lower or in_upper:
                        deletions.append(index)
        deletions = list(set(deletions))
        for index in sorted(deletions, reverse=True):
            del(self._positions[index])

    def get_mass(self):
        if self._masses is not None:
            return np.sum(self._masses)

    def get_rdf_ase(self, rmax, nbins, distance_matrix=None):
        ''' Adapted this function from ase.ga.utilities to use mic=True '''
        dm = distance_matrix 
        if dm is None: 
            dm = self.get_all_distances(mic=True)
        rdf = np.zeros(nbins + 1) 
        dr = float(rmax / nbins) 
        for i in range(len(self)): 
            for j in range(i + 1, len(self)): 
                rij = dm[i][j] 
                index = int(math.ceil(rij / dr)) 
                if index <= nbins: 
                    rdf[index] += 1 

        # Normalize 
        phi = len(self) / self.get_volume() 
        norm = 2.0 * math.pi * dr * phi * len(self) 

        dists = [0] 
        for i in range(1, nbins + 1): 
            rrr = (i - 0.5) * dr 
            dists.append(rrr) 
            rdf[i] /= (norm * ((rrr**2) + (dr**2) / 12.)) 

        return rdf, np.array(dists)


    def get_rdf(self, rmax, nbins):
        """ This function works for max 15000 particles """
        bins = np.linspace(0, rmax, nbins)
        distances = self.get_all_distances(mic=True)
        pairs=[]
        for cut, row in enumerate(distances[0:-1]):
            pairs += row.tolist()[cut+1:]
        rdf, edges = np.histogram(pairs, bins=bins, normed=False)
        # centralize the mesh
        r = 0.5 * (edges[1:] + edges[:-1])
        # volumes of the spherical shells betw. r and r+dr for each bin
        volumes = (4. / 3.) * np.pi * (np.power(edges[1:], 3) - np.power(edges[:-1], 3))
        # normalize by the number density
        ndens = len(pairs) / self.get_volume()
        rdf = rdf.astype(np.float) / (ndens * volumes)
        return np.insert(rdf, 0, 0.), np.insert(r, 0, 0.)


    def todict(self):
        """ Returns a morphology dictionary """
        morphology = {}
        morphology['positions'] = self.get_positions().tolist()
        morphology['masses'] = self.get_masses().tolist()
        morphology['pbc'] = self.get_pbc().tolist()
        morphology['cell'] = self.get_cell().tolist()
        return morphology


Morphology = Molecules




""" Methods for constructing structures with atomistic resolution """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'


def create_structure(**kwargs):
    """ Generic method to construct a structure with ASE from a dictionary """
    if kwargs['type'] == 'cluster':
        from ase import cluster
        from ase import Atoms
        module = getattr(cluster, kwargs['module'])
        method = getattr(module, kwargs['method']['name'])
        cluster = method(*kwargs['method']['args'], **kwargs['method']['kwargs'])
        return Atoms(cluster)

    if kwargs['type'] in ['slab', 'surface', 'bulk']:
        from ase import build
        method = getattr(build, kwargs['method']['name'])
        return method(*kwargs['method']['args'], **kwargs['method']['kwargs'])

    if kwargs['type'] == 'molecule':
        pass

    if kwargs['type'] == 'polymer':
        pass

    if kwargs['type'] == 'graphene':
        pass

    if kwargs['type'] == 'nanotube':
        pass

    if kwargs['type'] == 'molecular cluster':
        pass

    if kwargs['type'] == 'molecular slab':
        pass

    if kwargs['type'] == 'molecular bulk crystal':
        pass



""" ACFDT-RPA energy with VASP (contribution from Eric Hermes) """
from ase.calculators.vasp import Vasp
from shutil import copy

def get_acfdt_energy(species, common_params, nomega=10, acfdt_algo='RPAR',
                     acfdt_precfock='Fast', gamma_point=False, loptics=False):
    # step 1
    calc1 = Vasp(ediff=1e-8, **common_params)
    species.set_calculator(calc1)
    species.get_potential_energy()
    copy('OUTCAR', 'OUTCAR.1')
    copy('OSZICAR', 'OSZICAR.1')
    copy('vasprun.xml', 'vasprun.1.xml')

    # step 2
    calc2 = Vasp(algo='eigenval',
                 nelm=1,
                 lwave=False,
                 lhfcalc=True,
                 aexx=1.0,
                 precfock='Fast',
                 **common_params)

    species.set_calculator(calc2)
    try:
        species.get_potential_energy()
    except UnboundLocalError:
        pass
    copy('OUTCAR', 'OUTCAR.2')
    copy('OSZICAR', 'OSZICAR.2')
    copy('vasprun.xml', 'vasprun.2.xml')

    nbands = None
    eexx = None
    ehfc = None

    with open('OUTCAR.2', 'r') as f:
        for line in f:
            if line.startswith(" maximum number of plane-waves:"):
                nbands = int(line.split()[-1])
            elif line.startswith("  energy without entropy ="):
                eexx = float(line.split()[-1])
            elif line.startswith("  exchange ACFDT corr."):
                ehfc = float(line.split()[4])
                break
    if None in [nbands, eexx, ehfc]:
        raise ValueError('Could not parse HF OUTCAR!')

    if gamma_point:
        nbands = nbands * 2

    # step 3
    calc3 = Vasp(nbands=nbands,
                 algo='Exact',
                 nelm=1,
                 loptics=loptics,
                 **common_params)

    species.set_calculator(calc3)
    species.get_potential_energy()
    copy('OUTCAR', 'OUTCAR.3')
    copy('OSZICAR', 'OSZICAR.3')
    copy('vasprun.xml', 'vasprun.3.xml')

    # step 4
    calc4 = Vasp(nbands=nbands,
                 algo=acfdt_algo,
                 nomega=nomega,
                 precfock=acfdt_precfock,
                 **common_params)
    
    species.set_calculator(calc4)
    try:
        species.get_potential_energy()
    except UnboundLocalError:
        pass
    copy('OUTCAR', 'OUTCAR.4')
    copy('OSZICAR', 'OSZICAR.4')
    copy('vasprun.xml', 'vasprun.4.xml')

    with open('OUTCAR.4', 'r') as f:
        for line in f:
            if line.startswith("  converged value"):
                erpa = float(line.split()[2])

    return erpa, eexx, ehfc



__author__ = 'Martin Uhrin, Patrick Szmucer'

import numpy as np
from ase.atoms import Atoms
from ase.calculators.calculator import all_properties
from ase.calculators.singlepoint import SinglePointCalculator
from six import string_types


def filter_keys(keys_list, keys, omit_keys):
    '''Decides which keys to show given keys and omit_keys'''

    new_keys_list = list(keys_list)
    if omit_keys:
        if keys is not None:
            for key in keys_list:
                if key in keys:
                    new_keys_list.remove(key)
        else:
            # Omit all keys
            new_keys_list = []
    else:
        if keys is not None:
            for key in keys_list:
                if key not in keys:
                    new_keys_list.remove(key)
    return new_keys_list


def get_info_and_arrays(atoms, plain_arrays):
    """
    Extracts the info and arrays dictionaries from the Atoms object.
    If plain_arrays is True, numpy arrays are converted to lists.
    """
    info = {}
    arrays = {}
    for (key, value) in list(atoms.info.items()):
        key = key.lower()
        # TODO: all scalar values? test...
        if (isinstance(value, int) or isinstance(value, string_types) or
            isinstance(value, float) or isinstance(value, bool)):
            # Scalars
            info[key] = value
        else:
            # More complicated data structures
            arrays[key] = value

    skip_arrays = ['numbers', 'positions', 'species']
    for (key, value) in atoms.arrays.items():
        if key in skip_arrays:
            continue
        key = key.lower()
        arrays[key] = value

    if plain_arrays:
        for key, value in arrays.items():
            if value.__class__ == np.ndarray:
                arrays[key] = value.tolist()

    return info, arrays


def atoms2dict(atoms, plain_arrays=False):
    """
    Converts the Atoms object to a dictionary. If plain_arrays is True,
    numpy arrays are converted to lists.
    """
    d = {
        'numbers': atoms.numbers,
        'pbc': atoms.pbc,
        'cell': atoms.cell,
        'positions': atoms.positions}
    if atoms.has('magmoms'):
        d['initial_magmoms'] = atoms.get_initial_magnetic_moments()
    if atoms.has('charges'):
        d['initial_charges'] = atoms.get_initial_charges()
    if atoms.has('masses'):
        d['masses'] = atoms.get_masses()
    if atoms.has('tags'):
        d['tags'] = atoms.get_tags()
    if atoms.has('momenta'):
        d['momenta'] = atoms.get_momenta()
    if atoms.constraints:
        d['constraints'] = [c.todict() for c in atoms.constraints]
#    if atoms.calc is not None:
#        d['calculator'] = atoms.calc.name.lower()
#        d['calculator_parameters'] = atoms.calc.todict()
#        if len(atoms.calc.check_state(atoms)) == 0:
#            d.update(atoms.calc.results)

    if plain_arrays:
        for key, value in d.items():
            if value.__class__ == np.ndarray:
                d[key] = value.tolist()

    info, arrays = get_info_and_arrays(atoms, plain_arrays)
    d['info'] = info
    d['arrays'] = arrays
    return d


def dict2atoms(d, plain_arrays=False):
    """
    Converts a dictionary created with atoms2dict back to atoms.
    """
    atoms = Atoms(d['numbers'],
                  d['positions'],
                  cell=d['cell'],
                  pbc=d['pbc'],
                  magmoms=d.get('initial_magmoms'),
                  charges=d.get('initial_charges'),
                  tags=d.get('tags'),
                  masses=d.get('masses'),
                  momenta=d.get('momenta'),
                  constraint=d.get('constraints'))

    results = {}
    for prop in all_properties:
        if prop in d:
            results[prop] = d[prop]
    if results:
        atoms.calc = SinglePointCalculator(atoms, **results)
        atoms.calc.name = d['calculator']

    atoms.info['uid'] = d.get('uid')

    if 'arrays' in d:
        for key, value in d['arrays'].items():
            key = str(key)  # avoid unicode strings
            if plain_arrays:
                value = np.array(value)
            if value.dtype.kind == 'U':
                value = value.astype(str)
            try:
                atoms.new_array(key, value)
            except (TypeError, ValueError):
                atoms.info[key] = value
    if 'info' in d:
        for key, value in d['info'].items():
            atoms.info[str(key)] = value
    return atoms



""" A collection of functions to model the electrochemical potential """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'

import numpy as np
from ase import Atoms
from ase.io import read
from ase.build import add_adsorbate
from ase.constraints import FixAtoms
from ase.constraints import dict2constraint
from data import atoms2dict
from construct import create_structure
from surface import modify_surface_atom
from calculators import classmap
from structure import Structure

from parsl import python_app
from time import sleep
from datetime import datetime


@python_app
def MERGE(**kwargs):
    dic_merged = {}
    for key, value in kwargs.items():
        if isinstance(value, np.number):
            value = float(value)
        else:
            value = value
    return (kwargs)            


@python_app
def build_structure_generic(params):
    struct_dict = []
    sleep(2)
    return struct_dict

@python_app
def build_structure(params):
    """ construct a slab model and add adsorbates """
    struct_dict = []
    sleep(2)
    return struct_dict

@python_app
def structure_relaxation(params, struct_dict):
    now = datetime.now()
    print( now.strftime("%H:%M:%S"))
    """ perform structure relaxation with constraints """
    struct_dict = ['test']
    energy = []
    forces = []
    dipole = []
    sleep(2)
    return struct_dict, energy, forces, dipole

@python_app
def vibrations(params, struct_dict):
    """ perform normal mode analysis with constraints """
    # calculate properties
    entropy = 0
    partfun = 0
    vet = 0
    zpe = 0
    tst = 0
    minimum = 0
    vib_ene_real = 0
    sleep(2)
    return vib_ene_real, entropy, vet, partfun, zpe, tst, minimum

@python_app
def reaction_energies(props, reactions):
    """ calculate reaction energies """
    sleep(2)
    return reactions








""" building and manipulating periodic slab models """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'

from ase import Atoms
from ase.build import add_adsorbate
from ase.constraints import FixAtoms
from atomsutils import make_graph
from substrate import Substrate, SubstrateAdsorbateComplex

class Slab(Substrate):
    """ periodic slab structure """
    def __init__(self, params):
        super(Slab, self).__init__(**params)
        self.params = params

    def modify_atom(self, modification):
        from surface import modify_surface_atom
        modify_surface_atom(self.atoms, **modification)

    def set_constraints(self, constraints):
        """ slab-specific types of constraints """
        if constraints is None:
            return
        if all('layers' not in c for c in constraints):
            super(Slab, self).set_constraints(constraints)
            return
        for constraint in constraints:
            cmask = [atom.tag in constraint['layers'] for atom in self.atoms]
            self.atoms.set_constraint(FixAtoms(mask=cmask))

    def copy(self):
        """ create and return a copy of itself """
        return self.__class__(self.params)


class SlabAdsorbateComplex(SubstrateAdsorbateComplex):
    """ represent a slab-adsorbate complex """
    def __init__(self, slab, ads_params, constraints=None, vacuum=None):
        from copy import deepcopy
        ads_params_c = deepcopy(ads_params)
        self.atoms = slab.atoms.copy()
        for ads in ads_params_c:
            if isinstance(ads['adsorbate'], dict):
                ads['adsorbate'] = Atoms(**ads['adsorbate'])
            else:
                assert isinstance(ads['adsorbate'], Atoms)
            add_adsorbate(self.atoms, **ads)
        self.add_vacuum(vacuum)
        self.graph = make_graph(self.atoms)
        self.set_constraints(constraints)

    def set_constraints(self, constraints):
        pass


def get_all_complexes(slab, sites, ads_params, **kwargs):
    """ generate all possible slab-adsorbate complexes """
    from itertools import permutations
    complexes = []
    for stuple in permutations(sites, len(ads_params)):
        upd_pars = [{**ads, **site} for site, ads in zip(stuple, ads_params)]
        complexes.append((SlabAdsorbateComplex(slab, upd_pars, **kwargs), upd_pars))
    return complexes


def yield_all_complexes(slab, sites, ads_params, progress=False, **kwargs):
    """ generate all possible slab-adsorbate complexes - generator version """
    from itertools import permutations
    from tqdm import tqdm
    itera = permutations(sites, len(ads_params))
    total = len(list(permutations(sites, len(ads_params))))
    tqdm_itera = tqdm(itera, total=total, miniters=10) if progress else itera
    for stuple in tqdm_itera:
        upd_pars = [{**ads, **site} for site, ads in zip(stuple, ads_params)]
        yield (SlabAdsorbateComplex(slab, upd_pars, **kwargs), upd_pars)


def get_unique_complexes(complexes):
    """ filter out duplicated complexes and return the unique ones """
    unique = []
    for test in complexes:
        if not any(test[0].isomorphic(c[0]) for c in unique):
            unique.append(test)
    return unique



"""Perform simple workflow queries"""
import re
from prettytable import PrettyTable


WF_PROJECTION = {
    '$project': {
        'name': True,
        'metadata': True,
        'state': True,
        'nodes': True,
        'parent_links': True,
        'fws': True,
        'updated_on': True
    }
}


FW_PROJECTION = {
    '$project': {
        'fw_id': True,
        'name': True,
        'spec': True,
        'state': True,
        'updated_on': True,
        'updates': True,
        'launch_dir': True
    }
}


FW_PIPELINE = [
    {
        '$match': {
            '$expr': {
                '$in': ['$fw_id', '$$mynodes']
            }
        }
    },
    {
        '$project': {
            'last_launch_id': {
                '$slice': ['$launches', -1]
            },
            'fw_id': True,
            'name': True,
            'spec': True,
            'state': True,
            'updated_on': True
        }
    },
    {
        '$lookup': {
            'from': 'launches',
            'localField': 'last_launch_id',
            'foreignField': 'launch_id',
            'as': 'last_launch'
        }
    },
    {
        '$addFields': {
            'updates': {
                '$arrayElemAt': ['$last_launch.action.update_spec', 0]
            },
            'launch_dir': {
                '$arrayElemAt': ['$last_launch.launch_dir', 0]
            }
        }
    },
    FW_PROJECTION
]


CLEANUP_WFS = {'$match': {'fws': {'$not': {'$size': 0}}}}


def io_info(fw_dict, io_kind, storage):
    """
    Get tuples of data names and types from a firework, restricted to PyTask

    Args:
        fw_dict (dict): Firework dictionary
        io_kind (str): Either 'inputs' or 'outputs'
        storage (dict): Dictionary where the data has been stored
    Returns:
        ([tuple]): A list of tuples containing data names and data types
    """
    names = []
    for task in fw_dict['spec']['_tasks']:
        if task['_fw_name'] == 'PyTask':
            names.extend(task.get(io_kind, []))
    types = [type(storage.get(n)) for n in names]
    return list(zip(names, types))


def task_info(fw_dict):
    """
    Get info about dataflow tasks, restricted to PyTask
    Args:
        fw_dict (dict): Firework dictionary
    Returns:
        res ([dict]): A list of dictionaries with the following keys:
            name (str): the Python function called in the PyTask
            inputs (list, None): List of input data names
            outputs (list, None): List of output data names
    """
    res = []
    for task in fw_dict['spec']['_tasks']:
        if task['_fw_name'] == 'PyTask':
            info = {'name': task['func'], 'inputs': task.get('inputs'),
                    'outputs': task.get('outputs')}
            res.append(info)
    return res


class WFQuery(list):
    """
    Perform a query in fireworks launchpad

    Args:
        lpad (LaunchPad): a FireWorks LaunchPad object
        wf_query (dict): pymongo query for the workflows collection
        fw_query (dict): pymongo query for the fireworks collection
    """

    def __init__(self, lpad, wf_query={}, fw_query={}):
        query_wfs = {'$match': wf_query}
        query_fws = {'$lookup': {
            'from': 'fireworks',
            'let': {'mynodes': '$nodes'},
            'pipeline': [{'$match': fw_query}] + FW_PIPELINE,
            'as': 'fws'}}
        pipeline = [query_wfs, query_fws, WF_PROJECTION, CLEANUP_WFS]
        super().__init__(lpad.workflows.aggregate(pipeline))

    def get_wf_ids(self):
        """Return the firework ids, one for each worklfow"""
        return [next(fw['fw_id'] for fw in wf['fws']) for wf in self]

    def get_fw_ids(self):
        """Return the firework ids, complete list"""
        return [fw['fw_id'] for wf in self for fw in wf['fws']]

    def check_fw_ids(self, fw_ids=None):
        """
        If None return full list of fw_ids otherwise make a check

        Args:
            fw_ids ([int], None): list of firework ids
        Returns:
            retval ([int]): list of firework ids
        """
        my_fw_ids = self.get_fw_ids()
        if fw_ids is None:
            retval = my_fw_ids
        else:
            assert all(i in my_fw_ids for i in fw_ids), 'invalid fw_ids'
            retval = fw_ids
        return retval

    def get_wf_info(self, fw_ids=None):
        """
        Display a summary of workflows including specific nodes

        Args:
            fw_ids ([int], None): list of firework ids
        """
        my_fw_ids = self.check_fw_ids(fw_ids)
        fw_info = self.get_fw_info()
        for wfl in self:
            if any(i in wfl['nodes'] for i in my_fw_ids):
                templ = '\nworkflow name: {}\nworkflow state: {}\nupdated on: {}'
                print(templ.format(wfl['name'], wfl['state'], wfl['updated_on']))
                columns = fw_info[0].keys()
                table = PrettyTable(columns)
                for fwk in fw_info:
                    if fwk['fw_id'] in wfl['nodes']:
                        table.add_row([fwk[i] for i in columns])
                print(table)

    def get_fw_info(self, fw_ids=None):
        """
        Provide detailed information about specific nodes

        Args:
            fw_ids ([int], None): list of firework ids
        Returns:
            res ([dict]): a list of dictionaries with name, fw_id, state,
                          time stamp, parents, inputs and outputs
        """
        my_fw_ids = self.check_fw_ids(fw_ids)
        res = []
        for wfl in self:
            for fwk in wfl['fws']:
                if fwk['fw_id'] in my_fw_ids:
                    outputs = fwk.get('updates', {})
                    o_name_type = io_info(fwk, 'outputs', outputs)
                    i_name_type = io_info(fwk, 'inputs', fwk['spec'])
                    parents = wfl['parent_links'].get(str(fwk['fw_id']), [])
                    fwinfo_keys = ['name', 'fw_id', 'state', 'updated_on']
                    fwinfo = [fwk[k] for k in fwinfo_keys]
                    fwinfo += [parents, i_name_type, o_name_type]
                    fwinfo_keys += ['parents', 'inputs', 'outputs']
                    res.append(dict(zip(fwinfo_keys, fwinfo)))
        return res

    def get_i_names(self, fw_ids=None):
        """
        Return input data names in selected fireworks

        Args:
            fw_ids ([int], None): list of firework ids
        Returns:
            ({str}): a set of input data names in the specified fireworks
        """
        my_fw_ids = self.check_fw_ids(fw_ids)
        fws = self.get_fw_info(my_fw_ids)
        return set(i for fw in fws for i in fw['inputs'])

    def get_o_names(self, fw_ids=None):
        """
        Return output data names in selected fireworks
        Args:
            fw_ids ([int], None): list of firework ids
        Returns:
            ({str}): a set of output data names in the specified fireworks
        """
        my_fw_ids = self.check_fw_ids(fw_ids)
        fws = self.get_fw_info(my_fw_ids)
        return set(o for fw in fws for o in fw['outputs'])

    def get_task_info(self, fw_ids=None):
        """
        Return the dataflow from PyTask firetasks in selected fireworks

        Args:
            fw_ids ([int], None): list of firework ids
        Returns:
            res ([dict]): list of dictionaries containing the task info
        """
        my_fw_ids = self.check_fw_ids(fw_ids)
        res = []
        for wfl in self:
            for fwk in wfl['fws']:
                if fwk['fw_id'] in my_fw_ids:
                    res.append({'fw id': fwk['fw_id'], 'tasks': task_info(fwk)})
        return res

    def get_data(self, dname, io_kind):
        """
        Return the input/output data for a given data name

        Args:
            dname (str): data name
            io_kind (str): one of 'input' or 'output'
        Returns:
            res ([dict]): a list of dictionaries with the following keys:
                fw_id (int): firework ID
                dname (str): data name, the value is the data or None
        """
        assert io_kind in ['input', 'output']
        dct = 'spec' if io_kind == 'input' else 'updates'
        res = []
        for wfl in self:
            for fwk in wfl['fws']:
                if fwk.get(dct):
                    data = fwk[dct].get(dname)
                    if data:
                        res.append({'fw id': fwk['fw_id'], dname: data})
        return res

    def get_i_data(self, dname):
        """
        Return the input data for a given data name

        Args:
            dname (str): data name
        Returns:
            res ([dict]): a list of dictionaries with the following keys:
                fw_id (int): firework ID
                dname (str): data name, the value is the data or None
        """
        return self.get_data(dname, 'input')

    def get_o_data(self, dname):
        """
        Return the output data for a given data name

        Args:
            dname (str): data name
        Returns:
            res ([dict]): a list of dictionaries with the following keys:
                fw_id (int): firework ID
                dname (str): data name, the value is the data or None
        """
        return self.get_data(dname, 'output')

    def get_nodes_providing(self, dname, match='string'):
        """
        Return the ids of nodes providing a specified output

        Args:
            dname (str): output data name (or a regular expression)
            match (str): either 'string' for an exact string match or
                         'regex' for a regular expression match
        Returns:
            ([int]): a list of matching node ids
        """
        if match == 'string':
            test = lambda x: dname == x
        elif match == 'regex':
            test = lambda x: re.match(dname, x)
        fws = self.get_fw_info()
        return [f['fw_id'] for f in fws for o, _ in f['outputs'] if test(o)]



""" generic classes and functions to model substrate-adsorbate systems """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'

from ase import Atoms
from structure import Structure


class Substrate(Structure):
    """ generic substrate class """

    def get_gcn(self, sidxs):
        """ calculate the generalized coordination number, fcc only """
        cnmax = {1: 12, 2: 18, 3: 22, 4: 26}
        assert len(self.graph.subgraph(sidxs).clusters()) == 1
        nei = {n for idx in sidxs for n in self.graph.neighbors(idx) if n not in sidxs}
        cns = [self.graph.degree(idx) for idx in nei]
        return sum(cns)/cnmax[len(sidxs)]

    def get_cn(self, sidxs):
        """ calculate the coordination number """
        assert len(self.graph.subgraph(sidxs).clusters()) == 1
        nei = {n for idx in sidxs for n in self.graph.neighbors(idx) if n not in sidxs}
        return len(nei)


class SubstrateAdsorbateComplex(Structure):
    """ substrate-adsorbate complex superclass with generic methods """
    def __init__(self, substr, **kwargs):
        self.substr = substr.copy()
        self.adsorb = Atoms()
        atoms = self.substr.atoms + self.adsorb
        super(SubstrateAdsorbateComplex, self).__init__(atoms, **kwargs)



#from parsl import python_app
import time


#@python_app
def weighted_sleep(*args, **kwargs):
    if 'len_output' in kwargs:
        output_number = kwargs['len_output']
    else:
        output_number = 1
    if 'sleep_time' in kwargs:    
        sleep_time = kwargs['sleep_time']
    else:
        sleep_time = 1 
    id =  kwargs['id'] 
    #print('Node', id,' is Sleeping for', sleep_time, "seconds.")
    time.sleep(int(sleep_time))
    output_list = ['Null' for y in range(int(output_number))]
    return output_list



""" functions for modeling charge transport in organic semiconductors """
__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2016, Karlsruhe Institute of Technology'

from math import pi, sqrt, exp
import numpy as np
from ase.units import _hbar, _e, kB
from calculators import calculate, get_energy


def get_pairs(morphology, hopping_threshold):
    """ find pairs of molecules within a cutoff radius (hopping threshold) """
    centers_of_mass = morphology.get_positions()
    hopping_pairs = []
    hopping_sites = []
    hopping_indices = []
    for i in range(len(centers_of_mass)):
        for j in range(i):
            distance_vector = centers_of_mass[i]-centers_of_mass[j]
            distance = np.linalg.norm(distance_vector)
            if distance < hopping_threshold:
                hopping_pairs.append({
                    'distance': distance, 'left': i, 'right': j,
                    'distance vector': distance_vector.tolist()})
                if i not in hopping_indices:
                    hopping_sites.append({'index': i,
                    'center of mass': centers_of_mass[i].tolist()})
                    hopping_indices.append(i)
                if j not in hopping_indices:
                    hopping_sites.append({'index': j,
                    'center of mass': centers_of_mass[j].tolist()})
                    hopping_indices.append(j)
    return [hopping_sites, hopping_pairs]


def get_e_coupling(left_index, right_index, parameters, morphology,
                   hopping_sites):
    """ compute electronic coupling element for two hopping sites """
    assert 'carrier' in parameters, 'carrier type not defined'
    assert parameters['carrier'] in ('electron', 'hole'), (
        'carrier ', parameters['carrier'], ' not supported'
    )
    mo_type = 'HOMO' if parameters['carrier'] == 'hole' else 'LUMO'
    calc_parameters = parameters['calculator']['parameters']

    # structures
    pair = [morphology[left_index], morphology[right_index]]
    dimer = pair[0] + pair[1]

    # dimer
    calc = calculate(dimer, parameters['calculator']['name'], calc_parameters)
    # eigenvectors and eigen energies of the dimer
    [C, e] = calc.get_mos()
    # C^-1
    C_inv = np.linalg.inv(C)
    # overlap matrix and fock matrix: will use FC = SCe
    # overlap matrix of the dimer
    overlap = np.matmul(C_inv, C_inv.T)
    # Fock/Kohn-Sham matrix of the dimer
    fock = np.matmul(C_inv, np.matmul(e, C_inv.T))

    # first monomer
    ar_index = next(
        index for (index, hopping_site) in enumerate(hopping_sites)
        if hopping_site['index'] == left_index
    )
    phi_d_half = np.asarray(hopping_sites[ar_index][mo_type]['eigenvector'])

    # second monomer
    ar_index = next(
        index for (index, hopping_site) in enumerate(hopping_sites)
        if hopping_site['index'] == right_index
    )
    phi_a_half = np.asarray(hopping_sites[ar_index][mo_type]['eigenvector'])

    # electronic coupling J
    # auxiliary zero-valued half-vector
    phi_0_half = np.zeros((np.size(phi_d_half)))
    # donor orbital in the dimer AO basis
    phi_d = np.concatenate((phi_0_half, phi_d_half))
    # acceptor orbital in the dimer AO basis
    phi_a = np.concatenate((phi_a_half, phi_0_half))
    # site energy of donor
    H_dd = np.matmul(np.matmul(phi_d, fock), phi_d)
    # site energy of acceptor
    H_aa = np.matmul(np.matmul(phi_a, fock), phi_a)
    # charge transfer integral between donor and acceptor
    H_da = np.matmul(np.matmul(phi_d, fock), phi_a)
    # overlap between donor and acceptor
    S_da = np.matmul(np.matmul(phi_d, overlap), phi_a)
    # electronic coupling between the two monomers
    e_coupling = (H_da-0.5*(H_aa+H_dd)*S_da)/(1.0-S_da*S_da)

    return e_coupling


def reorg_energy(structure, parameters):
    """ calculate the reorganization energy for the neutral-ion transition """
    loc_parameters = parameters['calculator']['parameters'].copy()
    loc_structure = structure.copy()

    # only doublet charged state supported
    ion_multiplicity = 2
    calc_name = parameters['calculator']['name']

    # relaxation
    loc_parameters['task'] = 'optimize'
    e2_neutral = get_energy(loc_structure, calc_name, loc_parameters)

    # single point energy
    loc_parameters['multiplicity'] = ion_multiplicity
    loc_parameters['total charge'] = parameters['ion charge']
    loc_parameters['task'] = 'energy'
    e1_ion = get_energy(loc_structure, calc_name, loc_parameters)

    # relaxation
    loc_parameters['task'] = 'optimize'
    e2_ion = get_energy(loc_structure, calc_name, loc_parameters)

    # single point energy
    loc_parameters['multiplicity'] = parameters['multiplicity']
    loc_parameters['total charge'] = parameters['charge']
    loc_parameters['task'] = 'energy'
    e1_neutral = get_energy(loc_structure, calc_name, loc_parameters)

    # reorg energy
    return e1_neutral - e2_neutral + e1_ion - e2_ion


def marcus_rates(simulation):
    """ Marcus rates analysis """
    hbar = _hbar / _e
    kBT = kB*simulation['parameters']['temperature']
    lambda_t = simulation['properties']['reorganization energy']['value']
    hopping_sites = simulation['hopping sites']
    if 'positions' in simulation['morphology'].keys():
        morphology_size = len(simulation['morphology']['positions'])
    else:
        morphology_size = len(simulation['morphology']['molecules'])
    A = sqrt(pi/(lambda_t*kBT))/hbar
    B = 1.0/(4.0*lambda_t*kBT)
    energies = morphology_size*[None]
    for site in hopping_sites:
        energies[site['index']] = site['energy']
    for pair in simulation['hopping pairs']:
        deltaE = energies[pair['right']]-energies[pair['left']]
        pair['delta E'] = deltaE
        pair['rate'] = A*pair['coupling']**2*exp(-B*(lambda_t+deltaE)**2)


def abraham_miller_rates():
    pass


def mobility(simulation):
    """ implementation of the GEMM by Rodin et al. PRB 91, 155203 (2015) """
    # parameters
    dimension = 3 # effective spatial dimension of CT (n)
    C = 0.25 # fit parameter, for effective-medium model C=0.25, captures percolation limit
    M = 2 # connectivity parameter, the mean number of site neighbors with J^2 > 0

    energies = np.array([dic['energy'] for dic in simulation['hopping sites']])
    couplings = np.array([dic['coupling'] for dic in simulation['hopping pairs']])
    distances = np.array([dic['distance'] for dic in simulation['hopping pairs']])
    lambda_t = simulation['properties']['reorganization energy']['value']

    e_mean = np.mean(energies)
    sigma = np.std(energies)
    mean_coupling = np.mean((couplings**2)*(distances**2))
    kBT = kB*simulation['parameters']['temperature']
    beta = 1.0/kBT
    beta_sigma = beta*sigma
    beta_lambda = beta*lambda_t
    hbar = _hbar / _e
    mobility_0 = (
        (sqrt(pi)*beta_lambda**1.5)
        /(2.0*dimension*hbar)
        *exp(-beta_lambda/4.0)
        *M*mean_coupling/lambda_t**2
    )
    mobility_1 = (
        mobility_0
        *1.0/sqrt(1.0+beta_sigma**2/beta_lambda)
        *exp(-C*beta_sigma**2)
    )

    props = {}
    props['energy disorder'] = {'value': sigma, 'units': 'eV'}
    props['mean site energy'] = {'value': e_mean, 'units': 'eV'}
    props['mean electronic coupling'] = {'value': mean_coupling, 'units': 'eV^2 Ang^2'}
    props['mobility without energy disorder'] = {'value': mobility_0*1.e-16, 'units': 'cm^2/V.s'}
    props['mobility with energy disorder'] = {'value': mobility_1*1.e-16, 'units': 'cm^2/V.s'}
    simulation['properties'].update(props)

    return


def connected_component_analysis(morphology, hopping_pairs):
    from morphology import gyration_radius

    results = {}

    n_vs = len(morphology['positions'])
    positions = morphology['positions']
    gr = Graph(n_vs)
    gr.vs['label'] = range(n_vs)
    for pair in hopping_pairs:
        gr.add_edge(pair['left'], pair['right'])

    # radius of gyration for the whole morphology
    radius_of_gyration_morphology = gyration_radius(positions)

    rnet_rmor = []  # percolation ratio
    nnet = []       # number of networks
    nconnect = []   # averge number of connections per molecule

    couplings = []
    for pair in hopping_pairs:
        couplings.append(abs(pair['coupling']))
    ects = np.linspace(np.amin(couplings), np.amax(couplings), num=100)
    cinds = np.digitize(couplings, ects)
    cgroups = []
    for eindex, ect in enumerate(ects):
        grp = []
        for ind, cind in enumerate(cinds):
            if cind == eindex:
                pair = hopping_pairs[ind]
                grp.append((pair['left'], pair['right']))
        cgroups.append(grp)

    for eindex, ect in enumerate(ects):
        gr.delete_edges(cgroups[eindex])
        connected_components = gr.clusters()
        cluster_sizes = connected_components.sizes()
        max_cluster = cluster_sizes.index(max(cluster_sizes))
        cluster_vertices = connected_components[max_cluster]
        number_of_networks = len(connected_components)
        nnet.append(number_of_networks)
        # radius of gyration for the largest network
        com = [positions[v] for v in cluster_vertices]
        radius_of_gyration_network = gyration_radius(com)
        rnet_rmor.append(radius_of_gyration_network/radius_of_gyration_morphology)
        # average number of connections per molecule
        nconnect.append(np.sum(gr.degree())/n_vs)

    results['percolation ratio'] = rnet_rmor
    results['number of networks'] = nnet
    results['averge number of connections per molecule'] = nconnect
    results['electronic coupling thresholds'] = ects.tolist()

    return results


from graph import Graph

class MorphologyGraph(Graph):

    directed = False
    weighted = False

    def __init__(self, molecules, weights=None, **kwargs):
        if weights is not None:
            a_key = weights in ['rates', 'couplings', 'distance']
            a_number = isinstance(weights, (float, int))
            if not a_key and not a_number:
                raise ValueError('weights: ' + weights)
            self.hopping_pairs = kwargs['hopping_pairs']
            params = kwargs['parameters']
        if weights == 'rates':
            self.directed = True

        self.molecules = molecules
        self.positions = self.molecules.get_positions()

        Graph.__init__(self, len(self.positions), directed=self.directed)

        self.vs['label'] = range(len(self.positions))
        if weights == None:
            for pair in self.hopping_pairs:
                self.add_edge(pair['left'], pair['right'])
        if weights == 'couplings':
            for pair in self.hopping_pairs:
                self.add_edge(pair['left'], pair['right'], weight=abs(pair['coupling']))
        if weights == 'rates':
            for pair in self.hopping_pairs:
                # use the detailed balance principle
                kBT = kB*params['temperature']
                reverse_rate = pair['rate']*exp(pair['delta E']/kBT)
                self.add_edge(pair['left'], pair['right'], weight=pair['rate'])
                self.add_edge(pair['right'], pair['left'], weight=reverse_rate)
        if type(weights) in [float, int]:
            self.es['weight'] = weights
        if weights == 'distance':
            pass

        self.weighted = self.is_weighted()

    def get_maxflow(self, source, target, color=None):
        mflow = self.maxflow(source, target, capacity="weight")
        if color: mflow.es["color"] = color
        return mflow

    def get_maxflow_value(self, source, target):
        return self.get_maxflow(source, target).value

    def get_maxflow_cut(self, source, target):
        return self.get_maxflow(source, target).cut

    def get_maxflow_flow(self, source, target):
        return self.get_maxflow(source, target).flow

    def plot(self, graph_object, filename='morphgraph.png'):
        from igraph import plot
        plot(graph_object, layout=self.layout('fr'), target=filename)

    def get_total_maxflow(self, cutoff, direction=2, method='superweight',
                          **kwargs):

        self.add_vertex(name='super source', label='S')
        self.add_vertex(name='super target', label='T')
        sindex = next(vertex.index for vertex in self.vs
                      if vertex['name'] == 'super source')
        tindex = next(vertex.index for vertex in self.vs
                      if vertex['name'] == 'super target')

        positions = self.positions

        if method == 'superweight':
            minpos = np.amin(np.array(positions)[:, direction])
            maxpos = np.amax(np.array(positions)[:, direction])
            superweight = 2.0*np.max(self.es['weight'])
            for index, position in enumerate(positions):
                if position[direction] < minpos+cutoff:
                    self.add_edge('super source', index, weight=superweight)
                if position[direction] > maxpos-cutoff:
                    self.add_edge(index, 'super target', weight=superweight)
        if method == 'exponential':
            superweight = np.max(self.es['weight'])
            maxedge = self.es[np.argmax(self.es['weight'])].tuple
            mdistance = self.molecules.get_distance(maxedge[0], maxedge[1])
            print(maxedge, mdistance)
            minpos = np.amin(np.array(positions)[:, direction])-mdistance
            maxpos = np.amin(np.array(positions)[:, direction])+mdistance
            for index, position in enumerate(positions):
                pos = position[direction]
                weight = superweight*np.exp(-(pos-minpos)/cutoff)
                if weight > 1.e-3:
                    self.add_edge('super source', index, weight=weight)
                weight = superweight*np.exp(-(maxpos-pos)/cutoff)
                if weight > 1.e-3:
                    self.add_edge(index, 'super target', weight=weight)
        st_distance = self.shortest_paths(source=sindex, target=tindex)[0][0]
        assert st_distance > 2, ('distance betw. S and T: ' + str(st_distance)
            + '\nS neighbors: ' + str(self.neighbors(sindex))
            + '\nT neighbors: ' + str(self.neighbors(tindex)))
        return self.get_maxflow(sindex, tindex, **kwargs)


    def network_analysis(self, ec_threshold=None, draw_plots=False):
        from morphology import gyration_radius

        results = {}

        if ec_threshold:
            dellist = []
            for pair in self.hopping_pairs:
                if abs(pair['coupling']) < ec_threshold:
                    dellist.append((pair['left'], pair['right']))
            self.delete_edges(dellist)

        connected_components = self.clusters()
        cluster_sizes = connected_components.sizes()
        max_cluster = cluster_sizes.index(max(cluster_sizes))
        cluster_vertices = connected_components[max_cluster]
        number_of_networks = len(connected_components)

        if draw_plots:
            morphology_graph_file = 'morphology-graph.png'
            networks_graph_file = 'networks-graph.png'
            self.plot(filename=morphology_graph_file)
            connected_components.plot(filename=networks_graph_file)
            results['morphology graph file'] = morphology_graph_file
            results['networks graph file'] = networks_graph_file

        # RoG for the largest network
        com = [self.positions[v] for v in cluster_vertices]
        radius_of_gyration_network = gyration_radius(com)
        # RoG for the whole morphology
        radius_of_gyration_morphology = gyration_radius(self.positions)

        results['number of networks'] = number_of_networks
        results['network radius of gyration'] = radius_of_gyration_network
        results['morphology radius of gyration'] = radius_of_gyration_morphology

        return results



""" A simple command line interface to query FireWorks """
import json, yaml
from argparse import ArgumentParser
from fireworks.scripts.lpad_run import get_lp
from fireworks.fw_config import CONFIG_FILE_DIR
from dbquery import db_select

def query():
    m_description = 'A command line interface to query FireWorks'
    parser = ArgumentParser(description=m_description)
    parser.add_argument('-o', '--output', choices=['json', 'yaml'],
                        default='json', type=lambda s: s.lower(),
                        help='Set output format: json or yaml.')
    parser.add_argument('-c', '--config_dir',
                        help='path to configuration file (if -l unspecified)',
                        default=CONFIG_FILE_DIR)
    parser.add_argument('-l', '--launchpad_file',
                        help='path to launchpad file', default=None)
    parser.add_argument('-f', '--query_file', required=True,
                        help='path to query file')
    args = parser.parse_args()

    dumpf = yaml.dump if args.output == 'yaml' else json.dumps
    dumpkw = {'yaml': {'default_flow_style': False},
              'json': {'sort_keys': True, 'indent': 2,
                       'separators': (',', ': ')}}

    fmt = args.query_file.split('.')[-1]
    assert fmt in ['json', 'yaml']
    with open(args.query_file, 'r') as inp:
        query_dict = json.load(inp) if fmt == 'json' else yaml.safe_load(inp)
    assert query_dict is not None, 'empty query not allowed'
    print(dumpf(db_select(get_lp(args), **query_dict), **dumpkw[args.output]))



""" A collection of functions to model the electrochemical potential """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'

import numpy as np
from ase import Atoms
from ase.io import read
from ase.build import add_adsorbate
from ase.constraints import FixAtoms
from ase.constraints import dict2constraint
from data import atoms2dict
from construct import create_structure
from surface import modify_surface_atom
from calculators import classmap
from structure import Structure





def build_structure_generic(params):
    """ construct an atomic structure """
    return atoms2dict(Structure(**params['structure']).atoms)


def build_structure(params):
    """ construct a slab model and add adsorbates """
    # substrate structure
    if params.get('substrate') is not None:
        substr_params = params['substrate']
        if substr_params.get('input structure') is not None:
            # load the structure from params
            substr = substr_params['input structure']
        elif substr_params.get('input file') is not None:
            # read the structure from file
            substr = read(substr_params['input file'])
        else:
            # construct the structure
            substr = create_structure(**substr_params['construct'])

        # constraints
        constraints = None
        if substr_params.get('constrain all atoms'):
            constraints = [FixAtoms(mask=len(substr)*[True])]
        elif substr_params.get('constraints') is not None:
            cparams = substr_params['constraints']
            if isinstance(substr_params['constraints'], list):
                constraints = [dict2constraint(c) for c in cparams]
            else:
                if 'layers' in cparams:
                    cmask = [atom.tag in cparams['layers'] for atom in substr]
                    constraints = [FixAtoms(mask=cmask)]
                else:
                    constraints = [dict2constraint(cparams)]
        if constraints is not None:
            substr.set_constraint(constraints)
            # constr_list = [c.todict() for c in constraints]

        # magmoms
        magmoms = params['substrate'].get('initial magnetic moments')
        if isinstance(magmoms, list):
            substr.set_initial_magnetic_moments(magmoms)
        elif magmoms is not None:
            substr.set_initial_magnetic_moments(len(substr)*[magmoms])

        # modify surface atoms
        if 'modifications' in params['substrate']:
            for mod in params['substrate']['modifications']:
                modify_surface_atom(substr, **mod)

    struct = substr

    # adsorbate structure
    if 'adsorbates' in params and params['adsorbates'] is not None:
        for ads in params['adsorbates']:
            if isinstance(ads['adsorbate'], dict):
                ads['adsorbate'] = Atoms.from_dict(ads['adsorbate'])
            add_adsorbate(struct, **ads)

    struct_dict = atoms2dict(struct)
    return struct_dict


def structure_relaxation(params, struct_dict):
    """ perform structure relaxation with constraints """
    from data import Atoms

    calc_params = params['calculator']['parameters']
    calc_name = str(params['calculator']['name'])
    calc = classmap[calc_name](**calc_params)
    if calc_params.get('restart', False):
        struct = calc.get_atoms()
    else:
        struct = Atoms.from_dict(struct_dict)
        struct.set_calculator(calc)
    calc.calculate(struct)
    assert calc.converged

    struct_dict = atoms2dict(struct)
    energy = struct.get_potential_energy()
    forces = struct.get_forces()
    dipole = struct.get_dipole_moment()
    return struct_dict, energy, forces, dipole

def vibrational_entropy(energies, temperature):
    """ calculate vibrational entropy """
    from ase.units import kB
    beta = 1.0/(kB*temperature)
    entr = 0.0
    for ene in energies:
        exp_hob = np.exp(-ene*beta)
        entr += ene*beta*exp_hob/(1.0-exp_hob)-np.log(1.0-exp_hob)
    entr *= kB
    return entr


def vibrational_partition_function(energies, temperature):
    """ evaluate vibrational partition function """
    from ase.units import kB
    beta = 1.0/(kB*temperature)
    partf = 1
    for ene in energies:
        partf = partf /(1.0-np.exp(-ene*beta))
    return partf


def vibrations(params, struct_dict):
    """ perform normal mode analysis with constraints """

    # optionally extend the list of constraints
    if 'constraints' in params['normal mode analysis']:
        add_constr = params['normal mode analysis']['constraints']
        assert isinstance(add_constr, list)
        if ('constraints' in struct_dict and 
            isinstance(struct_dict['constraints'], list)):
            struct_dict['constraints'].extend(add_constr)
        else:
            struct_dict['constraints'] = add_constr

    struct = Atoms.from_dict(struct_dict)

    # test case of all-atom constraints
    fixatomsc = [c for c in struct.constraints if isinstance(c, FixAtoms)]
    constr = set([idx for c in fixatomsc for idx in c.get_indices()])
    if len(constr) == len(struct):
        return [], 0.0, 0.0, 1.0, 0.0, False, True

    calc_params = params['calculator']['parameters']
    calc_params.update(params['normal mode analysis']['parameters'])
    calc_name = str(params['calculator']['name'])

    calc = classmap[calc_name](**calc_params)
    struct.set_calculator(calc)
    calc.calculate(struct)
    assert calc.converged
    (vib_ene_real, vib_ene_imag) = calc.get_vibrational_energies()

    # calculate properties
    entropy = vibrational_entropy(vib_ene_real, params['temperature'])
    partfun = vibrational_partition_function(vib_ene_real, params['temperature'])
    vet = -params['temperature']*entropy
    zpe = 0.5*np.sum(vib_ene_real)
    tst = len(vib_ene_imag) == 1
    minimum = len(vib_ene_imag) == 0
    return vib_ene_real, entropy, vet, partfun, zpe, tst, minimum


def reaction_energies(props, reactions):
    """ calculate reaction energies """
    for key in props:
        if 'total_entropy_term' not in props[key]:
            props[key]['total_entropy_term'] = (
                props[key]['vibrational_entropy_term'] +
                props[key]['rotational_entropy_term'] +
                props[key]['translational_entropy_term']
            )
    for reac in reactions:
        for term in reac['equation']:
            assert props[term['species']]['energy_minimum']
        for key in ['energy', 'zero_point_energy', 'total_entropy_term']:
            reac[key] = 0.0
            for term in reac['equation']:
                reac[key] += term['coefficient']*props[term['species']][key]
        reac['zpe-corrected energy'] = reac['energy']+reac['zero_point_energy']
        reac['free energy'] = reac['zpe-corrected energy']+reac['total_entropy_term']
    return reactions


def get_ueffumax(reactions):
    """ calculate effective and critical potentials """
    fref = sum(r['free energy'] for r in reactions if r['electrons'] == 0)
    frst = sum(r['free energy'] for r in reactions)
    elec = sum(r['electrons'] for r in reactions)
    u0 = frst/elec
    ueff = u0 - fref/elec
    # sign is positive for cathode reactions
    sign = -np.sign(elec)
    # stopping criterion in case of convergence
    utol = 0.001
    # stopping criterion if umax is in invalid range
    ulim = 10.
    umax = ueff
    delta = ueff / 2.0
    while abs(delta) > utol:
        free_umax = [r['free energy']-r['electrons']*umax for r in reactions]
        if all(feu <= 0 for feu in free_umax):
            if sign*delta < 0:
                delta = -delta / 2.0
        else:
            if sign*delta > 0:
                delta = -delta / 2.0
        umax += delta
        if abs(umax) > ulim:
            umax = -sign*np.infty
            break
    assert sign*ueff > sign*umax
    return ueff, umax


def get_rds(reactions, u0=None):
    """ calculate the rate determining step """
    if u0 is None:
        frst = sum(r['free energy'] for r in reactions)
        elec = sum(r['electrons'] for r in reactions)
        u0 = frst/elec
    free_u0 = [r['free energy']-r['electrons']*u0 for r in reactions]
    rds_ind, rds_free_u0 = max(enumerate(free_u0), key=(lambda x: x[1]))
    return rds_free_u0, print_reaction(reactions[rds_ind])


def print_reaction(reaction):
    """ string representation of a reaction """
    lhs_terms = []
    rhs_terms = []
    for term in reaction['equation']:
        if term['coefficient'] < 0:
            term = str(abs(term['coefficient'])) + ' ' + term['species']
            lhs_terms.append(term)
        else:
            term = str(term['coefficient']) + ' ' + term['species']
            rhs_terms.append(term)
    return ' + '.join(lhs_terms) + ' -> ' + ' + '.join(rhs_terms)


def reverse_reactions(reactions):
    """ return the reverse reactions and their properties """
    result = []
    for r in reversed(reactions):
        react = {}
        for key in r:
            if key == 'equation':
                react['equation'] = []
                for t in reversed(r['equation']):
                    term = {}
                    term['coefficient'] = -t['coefficient']
                    term['species'] = t['species']
                    react['equation'].append(term)
            else:
                react[key] = -r[key]
        result.append(react)
    return result


def potentials_generic(props, reactions, u0=None):
    """ calculate electrochemical potentials and the rate determinig step """

    # calculate the free energies of the specified reactions
    reactions = reaction_energies(props, reactions)

    # effective reversible potential and critical potential
    ueff, umax = get_ueffumax(reactions)

    # rate determinging step at the standard reversible potential u_0
    delta_free_u0, rds = get_rds(reactions, u0)

    return reactions, rds, delta_free_u0, ueff, umax


def potentials(props):
    """ calculate electrochemical potentials for the O2 reduction reaction
        NOTE: deprecated, use potentials_generic() instead """

    u0 = 1.229 # in Volt

    # free energies of species
    free0 = {}
    for spec in ['*', '*o', '*oh', 'o2', 'h2', 'h2o']:
        assert props[spec]['energy_minimum']
        if 'total_entropy_term' in props[spec]:
            total_entropy_term = props[spec]['total_entropy_term']
        else:
            assert 'vibrational entropy term' in props[spec]
            assert 'rotational_entropy_term' in props[spec]
            assert 'translational_entropy_term' in props[spec]
            total_entropy_term = (
                props[spec]['vibrational entropy term'] +
                props[spec]['rotational_entropy_term'] +
                props[spec]['translational_entropy_term']
            )

        free0[spec] = (props[spec]['energy'] +
                       props[spec]['zero_point_energy'] +
                       total_entropy_term)

    # stages: educts and products
    stage1 = '* + 0.5 o2 + h2'
    stage2 = '*o + h2'
    stage3 = '*oh + 0.5 h2'
    stage4 = '* + h2o'

    # number of electrons at each stage
    electrons = {
        stage1: 2,
        stage2: 2,
        stage3: 1,
        stage4: 0
    }

    # free energies of each stage
    free0a = {
        stage1: free0['*'] + 0.5*free0['o2'] + free0['h2'],
        stage2: free0['*o'] + free0['h2'],
        stage3: free0['*oh'] + 0.5*free0['h2'],
        stage4: free0['*'] + free0['h2o']
    }

    # effective reversible potential
    ueff = (free0a[stage2]-free0a[stage4])/(electrons[stage2]-electrons[stage4])
    assert ueff > 0.0

    # critical potential
    tol = 0.001
    umax = ueff
    delta = ueff / 2.0
    while abs(delta) > tol:
        free_umax = {}
        for stage in free0a:
            free_umax[stage] = free0a[stage]-electrons[stage]*umax
        if (free_umax[stage2] >= free_umax[stage3] and
            free_umax[stage3] >= free_umax[stage4]):
            if delta < 0:
                delta = -delta / 2.0
        else:
            if delta > 0:
                delta = -delta / 2.0
        umax += delta

    # rate determinging stage
    free_u0 = {stage: free0a[stage]-electrons[stage]*u0 for stage in free0a}
    delta_free_u0 = {
        stage1+'->'+stage2: free_u0[stage2]-free_u0[stage1],
        stage2+'->'+stage3: free_u0[stage3]-free_u0[stage2],
        stage3+'->'+stage4: free_u0[stage4]-free_u0[stage3],
    }
    rds = max(delta_free_u0, key=delta_free_u0.get)

    return rds, delta_free_u0[rds], ueff, umax


def get_gas_props(props, temperature):
    """ calculate properties of all species: gas species and adsorbates """

    zpe = {
      'h2': 0.270, # eV, J. Phys. Chem. Ref. Data 36, 389 (2007)
      'o2': 0.098, # eV, J. Phys. Chem. Ref. Data 36, 389 (2007)
      'h2o': 0.558 # https://webbook.nist.gov/cgi/cbook.cgi?ID=C7732185
    }
    ent = {
      'h2': 1.3613e-3, # https://webbook.nist.gov/cgi/cbook.cgi?ID=C1333740
      'o2': 2.1370e-3, # https://webbook.nist.gov/cgi/cbook.cgi?ID=C7782447
      'h2o': 2.1669e-3 # https://webbook.nist.gov/cgi/cbook.cgi?ID=C7732185
    }
    orr_free = -4.916 # eV, free energy of O2 + 2H2 = 2H2O
    rprops = props
    for spe in rprops:
        if spe in ['h2', 'h2o']:
            rprops[spe]['energy_minimum'] = True
            rprops[spe]['zero_point_energy'] = zpe[spe]
            rprops[spe]['total_entropy_term'] = -temperature*ent[spe]
            rprops[spe]['free energy'] = (rprops[spe]['energy']+
                                          rprops[spe]['zero_point_energy']+
                                          rprops[spe]['total_entropy_term'])
        else:
            if 'total_entropy_term' not in rprops[spe]:
                rprops[spe]['total_entropy_term'] = (
                    rprops[spe]['vibrational entropy term']+
                    rprops[spe]['rotational_entropy_term']+
                    rprops[spe]['translational_entropy_term'])
            rprops[spe]['free energy'] = (rprops[spe]['energy']+
                                          rprops[spe]['zero_point_energy']+
                                          rprops[spe]['total_entropy_term'])
    # special method for o2
    h2o_free = rprops['h2o']['free energy']
    h2_free = rprops['h2']['free energy']
    o2_free = 2*(h2o_free-h2_free)-orr_free
    rprops['o2'] = {}
    rprops['o2']['free energy'] = o2_free
    rprops['o2']['total_entropy_term'] = -temperature*ent['o2']
    rprops['o2']['zero_point_energy'] = zpe['o2']
    rprops['o2']['energy'] = o2_free-(zpe['o2']-temperature*ent['o2'])
    rprops['o2']['energy_minimum'] = True
    return rprops



""" building and manipulating nanoparticle models """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'

import numpy as np
from ase import Atom
from ase import Atoms
from ase.neighborlist import NeighborList
from ase.constraints import FixAtoms
from igraph import Graph
from uuid import uuid4
from atomsutils import make_graph
from substrate import Substrate
from substrate import SubstrateAdsorbateComplex

class Nanoparticle(Substrate):
    """ a general nanoparticle class currently restricted to fcc crystalline
        nanoparticles
        TODO: Nanoparticle(Substrate) """
    types = [
        'free atom',
        'ontop adatom',
        'bridge adatom',
        'hollow adatom',
        '111 vertex',
        'surface',
        '100-110 edge, 111-111-100 vertex',
        '110 surface (top), 111-111 edge, 110-111 edge',
        '100 surface',
        '111 surface',
        'unknown',
        '110 surface (bottom)',
        'bulk'
    ]
    site_cns = {
        '4-fold hollow': [
            (5, 5, 5, 5), (6, 6, 6, 6), (5, 7, 7, 8), (6, 7, 7, 8),
            (7, 7, 8, 8), (8, 8, 8, 8)
        ],
        '3-fold hollow': [
            (5, 5, 5), (7, 7, 7), (4, 7, 7), (7, 7, 9), (7, 9, 9), (9, 9, 9),
            (6, 6, 9), (5, 7, 7), (6, 7, 9), (6, 6, 10), (6, 6, 11),
            (6, 7, 10), (6, 7, 11), (7, 9, 10), (7, 9, 11), (7, 10, 10),
            (7, 10, 11), (7, 11, 11), (9, 9, 10), (9, 10, 10), (9, 9, 11),
            (9, 10, 11), (9, 11, 11)
        ]
    }
    top_site_max_coord = 10
    neighborlist = []
    coordination = []
    atomtype = []
    ads_sites = []

    def __init__(self, atoms, cutoff=1.4):
        assert not any(atoms.pbc), 'periodic system detected'
        self.cutoff = cutoff
        self.atoms = atoms
        self.build_neighborslists()
        self.build_adsorption_sites()

    def copy(self):
        from copy import deepcopy
        new_np = self.__class__(self.atoms, cutoff=self.cutoff)
        new_np.atoms = self.atoms.copy()
        new_np.graph = self.graph.copy()
        new_np.atomtype = deepcopy(self.atomtype)
        new_np.ads_sites = deepcopy(self.ads_sites)
        new_np.coordination = deepcopy(self.coordination)
        new_np.neighborlist = deepcopy(self.neighborlist)
        return new_np

    def build_neighborslists(self):
        """ build a neighborlist, graph, asign coordination numbers """
        self.neighborlist = []
        self.coordination = []
        self.atomtype = []
        nl = NeighborList(
            cutoffs=len(self.atoms)*[self.cutoff],
            bothways=True,
            self_interaction=False
        )
        nl.update(self.atoms)
        for atom in self.atoms:
            neighb = nl.get_neighbors(atom.index)[0]
            coord = len(neighb)
            self.neighborlist.append(neighb.tolist())
            self.coordination.append(coord)
            self.atomtype.append(self.types[coord])
        self.graph = make_graph(self.atoms)

    def build_adsorption_sites(self):
        """ constructs all kinds of adsorption sites """
        self.ads_sites = []
        g = Graph(len(self.atoms))
        nl = NeighborList(
            cutoffs=len(self.atoms)*[self.cutoff],
            bothways=False,
            self_interaction=False
        )
        nl.update(self.atoms)
        neighborlist = [nl.get_neighbors(atom.index)[0].tolist()
                        for atom in self.atoms]
        for index, surfatom in enumerate(neighborlist):
            if self.atomtype[index] != 'bulk':
                for neighbor in surfatom:
                    if self.atomtype[neighbor] != 'bulk':
                        g.add_edge(index, neighbor)

        bulkatoms = [a.position for a in self.atoms
                     if self.atomtype[a.index] == 'bulk']
        self.bulkref = np.sum(bulkatoms, axis=0)/len(bulkatoms)

        self.build_on_top_sites(g)
        self.build_bridge_sites(g)

        for clen, kind in zip([3, 4], ['3-fold hollow', '4-fold hollow']):
            self.build_cyclic_sites(g, clen, kind)

        self.set_site_types()
        self.set_site_cns()
        self.set_site_gcns()

    def build_on_top_sites(self, g):
        for vertex in g.vs.indices:
            if self.coordination[vertex] < self.top_site_max_coord:
                # only neighbors from the surface
                neighbors = [self.atoms[ind] for ind in g.neighbors(vertex)]
                self.ads_sites.append(
                    AdsorptionSite([self.atoms[vertex]], 'on-top',
                                   bulkref=self.bulkref, neighbors=neighbors))

    def get_bridge_common_neighbors(self, g, vertex1, vertex2):
        """ find axial common neighbors of vertex1 and vertex2 """
        neighborset = {n for n in g.neighbors(vertex1) if n != vertex2}
        neighborset.intersection_update(
            {n for n in g.neighbors(vertex2) if n != vertex1})
        if len(neighborset) < 2:
            # the case of a bridge site on 100, 010 or 001 surfaces
            # any atoms that are common neighbors of vertex1 and vertex2
            neighborset = set(self.neighborlist[vertex1])
            neighborset.intersection_update(
                {n for n in self.neighborlist[vertex2] if n != vertex1})
        assert len(neighborset) > 0
        neighbors = [self.atoms[ind] for ind in neighborset]
        return neighbors

    def build_bridge_sites(self, g):
        for edge in g.es:
            (vertex1, vertex2) = edge.tuple
            neighbors = self.get_bridge_common_neighbors(g, vertex1, vertex2)
            if len(neighbors) < 3:
                atoms = [self.atoms[vertex1], self.atoms[vertex2]]
                self.ads_sites.append(
                    AdsorptionSite(atoms, 'bridge',
                                   bulkref=self.bulkref, neighbors=neighbors))

    def build_cyclic_sites(self, g, clen, kind):
        lst = [sorted(l) for l in g.get_subisomorphisms_vf2(Graph.Ring(clen))]
        sites = [list(x) for x in set(tuple(x) for x in lst)]
        for site in sites:
            atoms = [self.atoms[idx] for idx in site]
            cns = tuple(sorted([self.coordination[idx] for idx in site]))
            if any(cns == cn for cn in self.site_cns[kind]):
                self.ads_sites.append(
                    AdsorptionSite(atoms, kind, bulkref=self.bulkref,
                                   neighbors=self.get_site_neighbors(site)))

    def get_site_neighbors(self, site):
        neighborset = {n for idx in site for n in self.neighborlist[idx]}
        return [self.atoms[ind] for ind in neighborset]

    def set_constraints(self, level='all'):
        """ Sets FixAtoms constraints in the nanoparticle.
        Possible values for the 'level' argument are:
        - 'all': all atoms fixed (default)
        - 'none': no atoms fixed
        - 'bulk': all bulk atoms fixed
        - 1, 2, 3, ...: only these first layers are relaxed
        - 'sites': only atoms of occupied sites are relaxed
        - 'neighbors': only atoms and neighbors of occupied sites are relaxed
        """
        layers = None
        string_args = ['all', 'none', 'bulk', 'sites', 'neighbors']
        assert (isinstance(level, int) and level > 0) or level in string_args
        if level == 'all':
            self.atoms.set_constraint(FixAtoms(mask=len(self.atoms)*[True]))

        if level == 'bulk':
            layers = 1

        if isinstance(level, int):
            layers = level

        if layers is not None:
            g = self.graph.copy()
            # g.vs['id'] = range(len(g.vs))
            for _ in range(layers-1):
                surfs = [s.index for s in g.vs if g.degree(s.index) != 12]
                g.delete_vertices(surfs)
            bulks = [s['id'] for s in g.vs if g.degree(s.index) == 12]
            self.atoms.set_constraint(FixAtoms(indices=bulks))

        if level in ['sites', 'neighbors']:
            mask = len(self.atoms)*[True]            
            for site in self.ads_sites:
                if not site.vacant:
                    for atom in site.atoms:
                        mask[atom.index] = False
            if level == 'neighbors':
                for site in self.ads_sites:
                    if not site.vacant:
                        site_indices = [atom.index for atom in site.atoms]
                        for atom in self.get_site_neighbors(site_indices):
                            mask[atom.index] = False
            self.atoms.set_constraint(FixAtoms(mask=mask))

    def set_site_types(self):
        """ determine the type of adsorption sites for the given kinds """
        for site in self.ads_sites:
            if site.kind == '3-fold hollow':
                ngrs = [set(self.neighborlist[a.index]) for a in site.atoms]
                if len(ngrs[0] & ngrs[1] & ngrs[2]) == 1:
                    site.set_type('hcp')
                else:
                    site.set_type('fcc')
            else:
                site.set_type(None)

    def set_site_cns(self):
        """ determine and set the coordination numbers of site atoms """
        for site in self.ads_sites:
            idxs = [a.index for a in site.atoms]
            cns = tuple(sorted([self.coordination[idx] for idx in idxs]))
            site.set_coordination_numbers(cns)

    def set_site_gcns(self):
        """ set the generalized coordination numbers for all sites """
        for site in self.ads_sites:
            sidxs = [a.index for a in site.atoms]
            gcn = self.get_gcn(sidxs)
            site.set_gcn(gcn)

    def modify_atom(self, mod):
        pass


class AdsorptionSite:
    """ a class to represent an adsorption site """
    kinds = ['on-top', 'bridge', '3-fold hollow', '4-fold hollow']
    types = ['on-top 111', 'on-top 110', 'on-top 100', 'fcc', 'hcp',
             'long bridge', 'short bridge', 'trough', None]
    locations = ['vertex', 'edge', 'facet']
    atoms = None # a list of Atom objects
    neighbors = None # a list of Atom objects
    adsorbate = None # associated adsorbate object
    vacant = True
    surfdir = None
    site_id = None
    cns = None
    gcn = None

    def __init__(self, atoms, kind, neighbors=None, bulkref=None):
        assert kind in self.kinds
        self.atoms = atoms
        self.neighbors = neighbors
        self.kind = kind
        eps = np.sqrt(np.finfo(np.float).eps)

        atoms_positions = [a.position for a in atoms]
        centroid = 1./len(atoms)*np.sum(atoms_positions, axis=0)
        if kind in ['3-fold hollow', '4-fold hollow']:
            m0 = atoms[-1].position - centroid
            for atom in atoms:
                m1 = atom.position - centroid
                normal = np.cross(m0, m1)
                # test colinearity for 4-fold hollow site
                if np.linalg.norm(normal) > eps:
                    break

        if kind in ['bridge', 'on-top']:
            assert len(neighbors) > 0
            neigh_positions = [a.position for a in neighbors]
            ncentroid = 1./len(neighbors)*np.sum(neigh_positions, axis=0)
            normal = centroid - ncentroid
            # coplanar atoms and neighbors
            if np.linalg.norm(normal) < eps:
                ref = atoms[0] if kind == 'bridge' else neighbors[-1]
                m0 = ref.position - centroid
                for neighbor in neighbors:
                    m1 = neighbor.position - centroid
                    normal = np.cross(m0, m1)
                    # test colinearity
                    if np.linalg.norm(normal) > eps:
                        break

        assert(np.linalg.norm(normal) > eps)
        normal = normal/np.linalg.norm(normal)
        self.surfdir = np.dot(centroid - bulkref, normal)
        norm = abs(self.surfdir)
        if norm > eps:
            self.surfdir = self.surfdir / norm
        self.centroid = centroid
        self.normal = normal
        self.site_id = uuid4()

    def get_kind(self):
        return self.kind

    def set_type(self, stype):
        assert stype in self.types
        self.stype = stype

    def get_type(self):
        return self.stype

    def set_location(self, location):
        assert location in self.locations
        self.location = location

    def get_location(self):
        return self.location

    def set_coordination_numbers(self, cns):
        self.cns = cns

    def get_coordination_numbers(self):
        return self.cns

    def set_gcn(self, gcn):
        """ set the generalized coordination number """
        self.gcn = gcn

    def get_gcn(self):
        """ get the generalized coordination number """
        return self.gcn

    def add_adsorbate(self, zmat, alignment):
        assert self.vacant
        self.adsorbate = Adsorbate(self, zmat, alignment)
        self.vacant = False

    def get_atoms(self):
        return Atoms(self.atoms)

    def get_adsorbate(self):
        if not self.vacant:
            return self.adsorbate


class Adsorbate:
    """ a class to represent a mono-dentate adsorbate """
    site = None # associated site object
    graph = None # adsorbate's graph
    atoms = None # Atoms object containing the adsorbate atoms

    def __init__(self, site, zmat, align):
        """ """
        from zmatrix import Atoms
        assert site.vacant
        assert len(zmat) > 0
        assert ((len(zmat) == len(align) == 1) or
                (len(zmat) == 2 and len(align) == 3) or
                (len(zmat) > 2 and len(align) == 4))
        self.site = site
        site.adsorbate = self
        auxa = self._construct_aux_atoms(zmat, align)
        auxz = self._construct_aux_zmat(auxa, zmat, align)
        positions = auxa.get_positions().tolist()
        self.atoms = Atoms.from_zmat(auxz, positions=positions)
        del self.atoms[:2] # delete the auxiliary atoms
        self.graph = make_graph(self.atoms)
        self.site.vacant = False

    def _construct_aux_atoms(self, zmat, align):
        """ construct auxiliary atoms object """
        si = self.site
        a0 = si.neighbors[-1] if si.kind == 'on-top' else si.atoms[-1]
        auxa = Atoms([a0])
        auxa += Atom('X', position=si.centroid)
        a1pos = si.centroid + align[0]*si.normal*si.surfdir
        auxa += Atom(zmat[0]['symbol'].capitalize(), a1pos)
        return auxa

    def _construct_aux_zmat(self, auxa, zmat, align):
        """ construct auxiliary z-matrix """
        auxz = [{'label': 'auxa0', 'symbol': auxa[0].symbol}]
        auxz.append({'label': 'auxa1', 'symbol': 'X',
                     'distance': {
                         'reference': 'auxa0',
                         'value': auxa.get_distance(0, 1)
                    }})
        auxz.extend([line.copy() for line in zmat])
        auxz[2]['distance'] = {'reference': 'auxa1', 'value': align[0]}
        auxz[2]['angle'] = {
            'reference': 'auxa0',
            'value': auxa.get_angle(0, 1, 2)
        }
        if len(zmat) > 1:
            auxz[3]['angle'] = {'reference': 'auxa1', 'value': align[1]}
            auxz[3]['dihedral'] = {'reference': 'auxa0', 'value': align[2]}
        if len(zmat) > 2:
            auxz[4]['dihedral'] = {'reference': 'auxa1', 'value': align[3]}
        return auxz

    def get_atoms(self):
        return self.atoms.copy()


class NanoparticleAdsorbateComplex(SubstrateAdsorbateComplex):
    """ represent a cluster-adsorbate complex """
    def __init__(self, nanoparticle, site_ids, ads_params,
                 cluster_constraints='none', vacuum=None):
        np = nanoparticle.copy()
        self.ads_sites = [s for s in np.ads_sites if s.site_id in site_ids]
        zmats = [ads['zmat'] for ads in ads_params]
        aligns = [ads['alignment'] for ads in ads_params]
        for site, zmat, align in zip(self.ads_sites, zmats, aligns):
            site.add_adsorbate(zmat, align)
        np.set_constraints(level=cluster_constraints)
        adsorbates = [site.adsorbate for site in self.ads_sites]
        self.atoms = np.atoms
        for ads, par in zip(adsorbates, ads_params):
            magmoms = par.get('initial magnetic moments', None)
            ads.atoms.set_initial_magnetic_moments(magmoms)
            self.atoms += ads.atoms
        self.add_vacuum(vacuum)
        self.graph = make_graph(self.atoms)


def get_all_complexes(np, site_ids, ads_params, **kwargs):
    """ generate all possible nanoparticle-adsorbate complexes """
    from itertools import permutations
    return [NanoparticleAdsorbateComplex(np, stuple, ads_params, **kwargs)
            for stuple in permutations(site_ids, len(ads_params))]


def yield_all_complexes(np, site_ids, ads_params, progress=False, **kwargs):
    """ generate all possible nanoparticle-adsorbate complexes """
    from itertools import permutations
    from tqdm import tqdm
    itera = permutations(site_ids, len(ads_params))
    total = len(list(permutations(site_ids, len(ads_params))))
    tqdm_itera = tqdm(itera, total=total, miniters=10) if progress else itera
    for stuple in tqdm_itera:
        yield NanoparticleAdsorbateComplex(np, stuple, ads_params, **kwargs)



import os
import os.path
from pathlib import Path
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), "engine"))
import ipywidgets as widgets
import wfgenes.core.wgenerator
import json, yaml
from ipywidgets import Button, HBox, VBox, Layout
from IPython.display import HTML, display, clear_output, Markdown as md, JSON, Javascript, FileLink, Image, display_javascript, display_html, display_json
from ipyfilechooser import FileChooser
import time
from json2html import *
from fabric2 import Connection
import imp
import site

style = {'description_width': 'initial'}

HOME = str(Path.home())
wfgenes_path = os.path.join(site.getsitepackages()[0], 'wfgenes')
fig_path=os.path.join(wfgenes_path, 'fig','wfgenes_logo.png')



default_path = FileChooser('')
default_path.default_path = HOME
default_path.use_dir_icons = True
default_path.filter_pattern = []
default_path.title = 'Work Space'



update_default_path_button = widgets.Button(
    tooltip='Update Workspace',
    description='Update Workspace',
    style=style, 
    icon='folder')
update_default_path_button_output = widgets.Output()

generation_button = widgets.Button(description='WF Generation', icon='gears', tooltip='Workflow Generation')
generation_output_button = widgets.Output()

execution_button = widgets.Button(description='WF Execution', icon='play', tooltip= 'Workflow Execution')
execution_output_button = widgets.Output()

fireworks_button = widgets.Button(description='FireWorks', icon='fast-forward')
fireworks_output_button = widgets.Output()

dask_parsl_button = widgets.Button(description='Dask or Parsl', icon='fast-forward')
dask_parsl_output_button = widgets.Output()

simstack_button = widgets.Button(description='SimStack', icon='fast-forward')
simstack_output_button = widgets.Output()


#Define event when run wconfig button is clicked
def generation_button_clicked(b):
    clear_output()
    clear_wfgenes_output()
    with generation_output_button:
        import wfgenes.core.wfgenes_generator
        #imp.reload(wfgenes.core.wfgenes_generator)
        wfgenes.core.wfgenes_generator.display_generator()
        

def execution_button_clicked(b):
    clear_output()
    clear_wfgenes_output()
    with execution_output_button:
        display(HBox([fireworks_button, dask_parsl_button,simstack_button]), 
                fireworks_output_button, dask_parsl_output_button, simstack_output_button) 

def fireworks_execution_button_clicked(b):
    clear_output()
    #clear_wfgenes_output()
    clear_executor_output()
    import wfgenes.engine.wfengine_jupyter
    with fireworks_output_button:
        wfgenes.engine.wfengine_jupyter.display_wfengine()

def dask_parsl_execution_button_clicked(b):
    clear_output()
    clear_executor_output()
    with dask_parsl_output_button:
        import wfgenes.core.wfgenes_executor
        wfgenes.core.wfgenes_executor.display_executor()        
             
def clear_wfgenes_output():
    """ Clear top outputs """
    with generation_output_button:
        clear_output()
    with execution_output_button:
        clear_output()
        

def clear_executor_output():
    """ Clear top outputs """
    with fireworks_output_button:
        clear_output()
    with dask_parsl_output_button:
        clear_output()

def update_default_path_button_clicked(b):
    with update_default_path_button_output:
        clear_output()
        def_path=os.path.join(wfgenes_path, 'WS.yml')
        try:
            print(default_path.selected_path+' is set as your workspace')
        except TypeError as error:                 
            print('Choose a valid path')
    
        
generation_button.on_click(generation_button_clicked)
execution_button.on_click(execution_button_clicked)
fireworks_button.on_click(fireworks_execution_button_clicked)
dask_parsl_button.on_click(dask_parsl_execution_button_clicked)
update_default_path_button.on_click(update_default_path_button_clicked)



#Display buttons
display(Image(filename=fig_path, width = 150, height =150))
display(default_path)
display(update_default_path_button, update_default_path_button_output)
display(HBox([generation_button, execution_button]), generation_output_button, execution_output_button) 



""" wfGenes: Automatic workflow generator."""


__author__ = 'Mehdi Roozmeh'
__email__ = 'mehdi.roozmeh@kit.edu'
__copyright__ = 'Copyright 2020, Karlsruhe Institute of Technology'

import os
import os.path
import sys
from copy import deepcopy
from collections import OrderedDict
import json
import yaml
from wfgenes.core.builtin_wfgenes import merge_dic
from wfgenes.core.initial_wfgenes import BasewfGenes

ROOT_DIR = os.path.dirname(os.path.abspath(__file__)) 
class ParslwfGenes():
    """Generate Python wrapper for wConfig nodes orignized to run parallel with DASK delayed decorators"""
    def __init__(self, wfgenes_init, args):
        self.parsl_generation(wfgenes_init, args)

    def parsl_generation(self, wfgenes_init, args):  
        self.wfg_parsl = ""
        self.indent="    " 
        lazy_str = 'parsl_'
        parsl_path = os.path.join(wfgenes_init.workflow_path, 'Parsl')
        if not os.path.exists(parsl_path):
            os.makedirs(parsl_path)
        self.wrapper_path = os.path.join(parsl_path,
        wfgenes_init.interface_dict['workflow_name'] + '_parsl.py')
        parsl_import_path = os.path.join(ROOT_DIR, 'parsl_import.cfg')
        parsl_import = open(parsl_import_path, 'r')
        self.wfg_parsl += parsl_import.read()
        self.wfg_parsl +='\n'
              
        for i in range(wfgenes_init.routine_number):
            for j in range(wfgenes_init.subroutine_number[i]):
                # Cache variables inside loop
                func_file = wfgenes_init.func[i][j][0]
                func_module = wfgenes_init.func[i][j][1]
                func_nodup  =  wfgenes_init.func_nodup[i][j][0]
                func_global_nodup= wfgenes_init.func_global_nodup[i][j][0] 
                if func_file != 'BUILTIN' and func_nodup != 'duplicate' and func_global_nodup != 'global_duplicate':
                    self.wfg_parsl += 'from ' + str(func_file) + ' import ' + str(func_module) + ' as __' + str(func_module) +'\n'
        
        for i in range(wfgenes_init.routine_number):
            for j in range(wfgenes_init.subroutine_number[i]):
                # Cache variables inside loop
                func_file = wfgenes_init.func[i][j][0]
                func_module = wfgenes_init.func[i][j][1]
                func_nodup  =  wfgenes_init.func_nodup[i][j][0]
                func_global_nodup= wfgenes_init.func_global_nodup[i][j][0] 
                if func_file != 'BUILTIN' and func_nodup != 'duplicate' and func_global_nodup != 'global_duplicate':
                    self.wfg_parsl += str(func_module) + ' = python_app(__' + str(func_module) + ')' +'\n'
        

        if args['embedded'] == False:
            executor_arg_parser_path = os.path.join(ROOT_DIR, 'executor_arg_parser.cfg')
            executor_arg_parser = open(executor_arg_parser_path, 'r')
            self.wfg_parsl += self.indent
            self.wfg_parsl += executor_arg_parser.read()
        else:
            pass

        if args['embedded'] == True:
            self.wfg_parsl += "\n" + "def main(**resource):\n\n"
            
        else:
            self.wfg_parsl += "\nif __name__ == '__main__':\n\n" 
            self.wfg_parsl += "\n"+self.indent+"resource = wfgenes_argparser(args)\n\n"
                        
        self.wfg_parsl += "\n"+ self.indent+"start_time = time.time()"

        parsl_config_path = os.path.join(ROOT_DIR, 'parsl_config.cfg')
        parsl_config = open(parsl_config_path, 'r')
        self.wfg_parsl += self.indent
        self.wfg_parsl += parsl_config.read()

        for i in range(wfgenes_init.routine_number):
            for j in range(wfgenes_init.subroutine_number[i]):
                # Cache variables inside loop j loop
                self.func_builtin = 'NORMALFUNCTION'
                func_file = wfgenes_init.func[i][j][0]
                func_module = wfgenes_init.func[i][j][1]
                split_array = 'NULL'  # The varaiable which FOREACH iterates over
                zip_inputs = []
                if len(wfgenes_init.func[i][j]) > 2:
                    if wfgenes_init.func[i][j][2] == 'FOREACH':
                        self.func_builtin = 'FOREACH'
                        split_array = wfgenes_init.func[i][j][3]
                        chunk_size = str(wfgenes_init.func[i][j][4])
                        index_split = ''
                        
                        if len(wfgenes_init.func[i][j]) > 5:
                            for elements in wfgenes_init.func[i][j][5]['zip_inputs']:
                                zip_inputs.append(elements)    
                        
                        len_inputs = len(wfgenes_init.inputs[i][j])
                        for k in range(len_inputs):
                            if wfgenes_init.inputs[i][j][k] == split_array and self.func_builtin == 'FOREACH':
                                index_split = wfgenes_init.inputs_links[i][j][k][2]
                        
                func_global_nodup_0= wfgenes_init.func_global_nodup[i][j][0]
                len_inputs = len(wfgenes_init.inputs[i][j])
                len_outputs=len(wfgenes_init.outputs[i][j])
                for k in range(len_inputs):
                    # Cache variables inside loop k loop
                    input_k = wfgenes_init.inputs[i][j][k]
                    inputs_nodup_k = wfgenes_init.inputs_nodup[i][j][k]
                    inputs_global_nodup_k = wfgenes_init.inputs_global_nodup[i][j][k]
                    inputs_no_locdep_k = wfgenes_init.inputs_no_locdep[i][j][k][0]
                    input_gdependent = wfgenes_init.inputs_links[i][j][k][0].isnumeric()
                          
                    if inputs_global_nodup_k != 'global_duplicate' and inputs_nodup_k!= 'duplicate' and inputs_no_locdep_k!= 'inner_dependent' and input_gdependent == False:
                        # Load wfgenes_init.inputs
                        self.wfg_parsl += '\n'+ self.indent +'#Read Input #' + str(k + 1)  + 'from subroutine #' + str(j + 1) + ' in routine #' + str(i) + '\n'
                        self.wfg_parsl += self.indent+"yaml_stream = open('" + input_k  + ".yaml', 'r')\n"
                        self.wfg_parsl += self.indent + input_k  + " = yaml.load(yaml_stream, Loader=yaml.Loader)\n\n"
        self.wfg_parsl += "\n"+ self.indent+"end_memtime = time.time()"
                            
        # Start of computation phase                    
        
        func_done = []
        step_simulation = 1
        wfgenes_init.wfgenes_scheduler(func_done)
        previously_computed = [] # Keep track of launched task to avoid repeated computation
        steps_width = []
        while wfgenes_init.func_waiting == True:        
            self.wfg_parsl += '\n\n'+self.indent+'#### Start Step #'+ str(step_simulation)
            width_counter = 0
            for i in range(wfgenes_init.routine_number):
                for j in range(wfgenes_init.subroutine_number[i]):
                    # Cache variables inside loop j loop
                    self.func_builtin = ''
                    func_file = wfgenes_init.func[i][j][0]
                    self.func_module = wfgenes_init.func[i][j][1]                             
                    func_global_nodup_0= wfgenes_init.func_global_nodup[i][j][0]
                    len_inputs = len(wfgenes_init.inputs[i][j])
                    len_outputs=len(wfgenes_init.outputs[i][j])
        
                    if wfgenes_init.foreach_output[i][j][0] != 'null':
                        # Dependent foreach
                        output_number =  len(wfgenes_init.outputs[wfgenes_init.foreach_output[i][j][1]][wfgenes_init.foreach_output[i][j][2]])

                    elif wfgenes_init.foreach_output[i][j][1] == 'null':
                        # independent foreach
                        output_number = len_outputs



                    if func_global_nodup_0 == 'global_duplicate': 
                        func_suffix = '_id' + str(i)
                    else:
                        func_suffix = ''     
                    self.split_array = 'NULL'  # The varaiable which FOREACH iterates over
                    self.zip_inputs = []
                    coordinate = [[i,j]]  # Routine and subroutine indexes
                    if len(wfgenes_init.func[i][j]) > 2:
                        if wfgenes_init.func[i][j][2] == 'FOREACH':
                            self.func_builtin = 'FOREACH'
                            self.split_array = wfgenes_init.func[i][j][3]
                            self.chunk_size = str(wfgenes_init.func[i][j][4])
                            self.index_split = ''
                            
                            if len(wfgenes_init.func[i][j]) > 5:
                                for elements in wfgenes_init.func[i][j][5]['zip_inputs']:
                                    self.zip_inputs.append(elements)    
                            
                            for k in range(len_inputs):
                                if wfgenes_init.inputs[i][j][k] == self.split_array and self.func_builtin == 'FOREACH':
                                    self.index_split = wfgenes_init.inputs_links[i][j][k][2]
                            
                            if self.chunk_size == 'full' and wfgenes_init.foreach_output[i][j][0] == 'null' :
                                self.chunk_size = 'len('+self.split_array+')'
                                self.step= '1'
                            elif self.chunk_size == 'full' and wfgenes_init.foreach_output[i][j][0] != 'null' :
                                self.chunk_size = 'len('+ lazy_str + wfgenes_init.foreach_output[i][j][0]+')'
                                self.step= '1'
                            elif self.chunk_size != 'full' and wfgenes_init.foreach_output[i][j][0] == 'null' :   
                                self.step= ' int(len('+self.split_array+')/'+self.chunk_size+')'
                            elif self.chunk_size != 'full' and wfgenes_init.foreach_output[i][j][0] != 'null' :
                                if output_number > 1:
                                    self.step = ' int(len('+ lazy_str +wfgenes_init.foreach_output[i][j][0]+'['+ self.index_split +'])/'+self.chunk_size+')'
                                elif output_number == 1:
                                    self.step = ' int(len('+ lazy_str +wfgenes_init.foreach_output[i][j][0]+')/'+self.chunk_size+')'


                            if wfgenes_init.foreach_output[i][j][0] == 'null':
                                foreach_len = 'range(0 , len('+self.split_array+'),'+self.step+')'
                            elif wfgenes_init.foreach_output[i][j][0] != 'null':
                                if output_number > 1:
                                    foreach_len = 'range(0, len('+ lazy_str +wfgenes_init.foreach_output[i][j][0]+'['+ self.index_split +'])'+','+self.step+ ')'         
                                elif output_number == 1:
                                    foreach_len = 'range(0, len('+ lazy_str +wfgenes_init.foreach_output[i][j][0]+')'+','+self.step+ ')'

                    if len(wfgenes_init.func_depid[i][j]) == 1 and coordinate[0] not in func_done:
                        if  self.func_module != 'MERGE':
                            wfgenes_init.func[i][j][1] = wfgenes_init.func[i][j][1] + func_suffix
                        # check if func is Ready and not computed already 
                        self.wfg_parsl += '\n'+self.indent+'# Call subroutine #' + str(j + 1) + ' from routine #' +str(i)+ '\n'
                        self.wfg_parsl += self.indent    
                    # Call functions
                        if self.func_module != 'MERGE' and self.func_builtin != 'FOREACH':
                            self.wfg_parsl += '\n'+self.indent+lazy_str+self.func_module+ func_suffix + ' = ' + self.func_module + '('
                            func_done.append([i ,j]) # Append implemented module
                            width_counter +=1
                            func_waiting = False     # Temporarily Set func_waiting  to False 
                            input_slice = ''   
                        elif self.func_module == 'MERGE' and self.func_builtin != 'FOREACH':
                            self.wfg_parsl += '\n'+self.indent +lazy_str+ wfgenes_init.outputs[i][j][0] + ' = ' +self.func_module+'('
                            func_done.append([i ,j]) # Append implemented module
                            width_counter +=1
                            func_waiting = False     # Temporarily Set func_waiting  to False 
                            input_slice = ''
                        elif self.func_builtin == 'FOREACH':
                            # foreach implementation part 1/2 
                            self.wfg_parsl += '\n'+self.indent+lazy_str+self.func_module+func_suffix+'= []'
                            self.wfg_parsl += '\n'+self.indent+lazy_str+self.func_module+func_suffix+'_future= []'
                            for k in range(len(wfgenes_init.foreach_output[i][j])):
                                if wfgenes_init.foreach_output[i][j][0] != 'null' and len(wfgenes_init.func[wfgenes_init.foreach_output[i][j][1]][wfgenes_init.foreach_output[i][j][2]]) == 2: 
                                    # Check if foreach computation depends on a lazy object
                                    compute_now = wfgenes_init.foreach_output[i][j][0] + func_suffix
                                    if compute_now not in previously_computed:
                                        previously_computed.append(compute_now) # Append computed lazy task- Avoid extra computation 
                                        self.wfg_parsl += '\n'+self.indent+lazy_str + compute_now+  '='+ lazy_str + compute_now+ '.result()' # Compute lazy objects   
                            self.wfg_parsl += '\n'+self.indent + 'for i in ' + foreach_len + ' :'
                            self.wfg_parsl += '\n'+self.indent+self.indent+lazy_str+self.func_module+func_suffix+ '_foreach=' +self.func_module + '('
                            # end of foreach implementation part 1/2
                        self.write_arguments(wfgenes_init, lazy_str, i, j)  # write argument list                                  
                        if self.func_builtin == 'FOREACH': # foreach implementation Part 2/2
                            self.wfg_parsl += '\n'+self.indent+self.indent+lazy_str+self.func_module+func_suffix+'_future.append('+lazy_str+self.func_module+func_suffix+'_foreach)'
                            previously_computed.append(self.func_module)
                            self.wfg_parsl += '\n'+self.indent + 'for i in range(len('+lazy_str+self.func_module+func_suffix+'_future)):'
                            self.wfg_parsl += '\n'+self.indent + self.indent+lazy_str+self.func_module+func_suffix+'.append('+lazy_str+self.func_module+func_suffix+'_future[i].result())'
                            if len(wfgenes_init.outputs[i][j]) > 1:
                                self.wfg_parsl += '\n'+self.indent+lazy_str + self.func_module + func_suffix + '= flat_tuple( '+ lazy_str + self.func_module + func_suffix + ', ' + str(len(wfgenes_init.outputs[i][j])) +')'
                            elif len(wfgenes_init.outputs[i][j]) == 1:    
                                self.wfg_parsl += '\n'+self.indent+lazy_str + self.func_module + func_suffix +  '= flat_list( '+ lazy_str + self.func_module + func_suffix + ')'
                            func_done.append([i ,j]) ## Append implemented module
                            width_counter +=1
            self.wfg_parsl += '\n'+self.indent+'#End of step ' + str(step_simulation) + ' with the width ' + str(width_counter)
            step_simulation = step_simulation + 1
            steps_width.append(width_counter)
            wfgenes_init.wfgenes_scheduler(func_done)

        if wfgenes_init.func[func_done[-1][0]][func_done[-1][1]][1] != 'MERGE':
            self.wfg_parsl += '\n'+self.indent+'print('+lazy_str+wfgenes_init.func[func_done[-1][0]][func_done[-1][1]][1]+'.result())'
        else:
            self.wfg_parsl += '\n'+self.indent+'print('+lazy_str+wfgenes_init.outputs[func_done[-1][0]][func_done[-1][1]][0]+ func_suffix+'.result())'                
        self.wfg_parsl += "\n"+ self.indent+"end_time = time.time()"
        self.wfg_parsl += "\n"+ self.indent+"total_time = end_time - start_time"
        self.wfg_parsl += "\n"+ self.indent+"total_memtime = end_memtime - start_time"
        self.wfg_parsl += "\n"+ self.indent+"print(' The total time is' , round(total_time, 2), 'seconds')"
        self.wfg_parsl += "\n"+ self.indent+"print(' The total memory load time is' , round(total_memtime, 2), 'seconds')"
        self.wfg_parsl += "\n"+self.indent+"print('PARSL based workflow')"
        self.wfg_parsl += "\n"+ self.indent + "#The model has "+ str(len(steps_width)) + " step with the width of "    
        for step in steps_width:
            self.wfg_parsl += str(step) + ", "
        with open(self.wrapper_path, 'w') as file:
                file.write(self.wfg_parsl)
        cmd = 'autopep8 --in-place --aggressive --aggressive '+ self.wrapper_path      
        os.system(cmd)                 



    
    def write_arguments(self, wfgenes_init, lazy_str, i , j):
        """ Writing argument list for a module call"""
        len_inputs = len(wfgenes_init.inputs[i][j])
        if bool(wfgenes_init.kwargs[i][j]) and wfgenes_init.kwargs[i][j] != 'null':
            kwargs = wfgenes_init.kwargs[i][j]
            last_key = list(kwargs)[-1]
            kwargs_string = ''
            for key, value in kwargs.items():
                kwargs_string +=  key + " = '" + str(value) + "'"      
                if key != last_key:
                    kwargs_string += ', '
                else:
                    kwargs_string += ')'
        for k in range(len_inputs):
            # Writing arguments of each function
            # Cache variables inside loop k loop
            inputs = wfgenes_init.inputs_py[i][j][k]
            inputs_nodup_k= wfgenes_init.inputs_nodup[i][j][k]
            inputs_no_locdep_k= wfgenes_init.inputs_no_locdep[i][j][k][0]
            if len(wfgenes_init.inputs_links[i][j][k]) == 3:
                input_gdependent = wfgenes_init.inputs_links[i][j][k][0].isnumeric()
            else:
                input_gdependent = False    
            argument_key = ''
            func_gloab = ''
            func_loc = ''
            func_loc_dep = ''
            func_gloab_dep = ''
            parsl_barrier = ''
            slice = '' # Slice only will be set for FOREACH method       
            if self.func_module == 'MERGE':
                argument_key = wfgenes_init.inputs_locname[i][j][k]+'=' # This prefix is set only for merge function
            
            if len(wfgenes_init.func[i][j]) > 2:
                if wfgenes_init.func[i][j][2] == 'FOREACH':
                    if inputs == self.split_array or inputs in self.zip_inputs:
                        slice = '[i:i+'+self.step+']'    
            if inputs_no_locdep_k== 'inner_dependent':
                input_index = wfgenes_init.inputs_no_locdep[i][j][k][2]
                func_loc = wfgenes_init.func[i][int(wfgenes_init.inputs_no_locdep[i][j][k][1])][1]
                func_loc_dep = func_loc 
                output_number = len(wfgenes_init.outputs[i][int(wfgenes_init.inputs_no_locdep[i][j][k][1])])
                if self.func_builtin == 'FOREACH' and inputs == self.split_array :
                    # Rewrite the value if it is split array
                    func_loc_dep = wfgenes_init.foreach_output[i][j][0]
                if self.func_builtin != 'FOREACH':
                    parsl_barrier = '.result()' 
                
            if input_gdependent == True:
                input_index = wfgenes_init.inputs_links[i][j][k][2]          
                func_gloab = wfgenes_init.func[int(wfgenes_init.inputs_links[i][j][k][0])][int(wfgenes_init.inputs_links[i][j][k][1])][1]
                func_gloab_dep = func_gloab 
                output_number= len(wfgenes_init.outputs[int(wfgenes_init.inputs_links[i][j][k][0])][int(wfgenes_init.inputs_links[i][j][k][1])])
                if self.func_builtin == 'FOREACH' and inputs == self.split_array :
                    # Rewrite the value if it is split arry
                    parsl_barrier = '' 
                    func_gloab_dep = wfgenes_init.foreach_output[i][j][0]
                elif self.func_builtin != 'FOREACH':
                    parsl_barrier = '.result()' 
                    

            # Writing arguments - 4 different scenarios
            if inputs_no_locdep_k != 'inner_dependent' and input_gdependent == False :
                #1 No dependency at all
                self.wfg_parsl += argument_key+ inputs  + parsl_barrier+slice+', '   
            elif inputs_no_locdep_k== 'inner_dependent' and func_loc != 'MERGE' :
                #2 Inner-dependent for users function
                self.wfg_parsl += argument_key+lazy_str + func_loc_dep  
                if output_number > 1:
                    self.wfg_parsl += parsl_barrier+'[' + input_index + '] ' + slice + ', '
                elif output_number == 1:
                    self.wfg_parsl += slice + ', '
            elif input_gdependent == True and func_gloab != 'MERGE' :
                #3 Gloab dependent for users function
                self.wfg_parsl += argument_key+lazy_str + func_gloab_dep  
                if output_number != 1:
                    self.wfg_parsl += parsl_barrier+'[' + input_index + ']'  + slice + ', '
                else:
                    self.wfg_parsl += slice + ', '
                        
            elif (func_loc == 'MERGE' or func_gloab == 'MERGE') and (input_gdependent == True or inputs_no_locdep_k== 'inner_dependent')  :
                #4 Dependent to Merge function
                    self.wfg_parsl += argument_key+lazy_str + inputs + ', '    
         
        if not bool(wfgenes_init.kwargs[i][j]) or wfgenes_init.kwargs[i][j] == 'null' : 
            if len_inputs !=0:
                self.wfg_parsl = self.wfg_parsl[:-2]
            self.wfg_parsl += ')'                 
        elif bool(wfgenes_init.kwargs[i][j]) and wfgenes_init.kwargs[i][j] != 'null' :
            self.wfg_parsl += kwargs_string

        

   
                            
                                
                                            
                    
                                    
                                        
                      

                                                                   
                   


                